%-------------------------------------------------------------------------------------
   Verification of Alternating Bit Protocol.
   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.ruu.nl


  The program is given below:

       PROG  ALT_BIT
       READ  wire, Sbit, buffer, Rbit, ack
       WRITE wire, Sbit, buffer, Rbit, ack
       INIT Sbit<>ack /\ Rbit=ack
       ASSIGN
          (IF Sbit = ack --> wire, Sbit := Exp, Next Sbit)           { Send }
          [] (buffer, Rbit := wire, Sbit)                            { Receive }
          [] (ack := Rbit)                                           { ACK }

  The specification is:

       !X. (J /\ wire=X) |--> (buffer=X) 
  
  for some invariant J.

  The variable `wire` is cotrolled by the Sender and it is free to update its 
  value as long as above specification is maintained. `buffer` is the copy of the 
  `wire` maintained by the Reciever.
 
  `Sbit` is the sender's alternating bit, and `Rbit` is the copy of `Sbit` 
   maintained by the reciever. `Ack` is the acknowledgement that the Reciever
   finally sends back to the sender.

   The function Next is assumed to have the following property:

          Next n <> n

-------------------------------------------------------------------------------------%
 
loadf `UU_UNITY` ;;
new_theory `ALT_BIT` ;;
load_library `string` ;;
new_parent `UPROG` ;;
Autoload `UPROG` ;;
loadf `UPROG_SUP` ;;
timer true ;;

let XVar = ":string" ;;
let XState = ":^XVar->num" ;;
let XExp = ":^XState->num" ;;
let XPred  = ":^XState->bool" ;;
let XAction = ":^XState->^XState->bool" ;;

let AB_Rd = new_definition (`Ab_Rd`,
   "AB_Rd = [`wire`; `Sbit`; `buffer`; `Rbit`; `ack`]") ;;

let AB_Wr = new_definition (`Ab_Wr`,
   "AB_Wr = [`wire`; `Sbit`; `buffer`; `Rbit`; `ack`]") ;;

let Send = new_definition (`Send`,
   "Send Exp Next = (\s:^XState. s `Sbit` = s `ack`) THEN
                      ((`wire` , `Sbit`) 
                      ASG2 
                      (Exp , (\s:^XState. Next (s `Sbit`))))") ;;

let Receive = new_definition (`Receive`,
   "Receive = (`buffer`,`Rbit`) 
              ASG2 
              ((\s:^XState. s `wire`) , (\s. s `Sbit`))") ;;

let ACK = new_definition (`ACK`,
   "ACK = `ack` ASG (\s:^XState. s `Rbit`)") ;;

let Init = new_definition (`Init`,
   "Init = \s:^XState. ~(s `Sbit` = s `ack`) /\  (s `Rbit` = s `ack`)") ;;

let ALT_BIT = new_definition (`ALT_BIT`,
   "ALT_BIT Exp Next = UPROG AB_Rd AB_Wr Init [Send Exp Next; Receive; ACK]") ;;

let ALT_BIT_defs = map snd (definitions `ALT_BIT`) ;;

%-----------------------------------------------------------------------------
  Whether ALT_BIT is a UNITY program can be checked using UNITY_DECOM_TAC
-----------------------------------------------------------------------------%

let AB_UNITY = prove 
  ("RC [Send Exp Next; Receive; ACK] AB_Rd 
       ==> UNITY (ALT_BIT Exp Next)", 
    REWRITE_TAC ALT_BIT_defs THEN UNITY_DECOM_TAC 5) ;;

%------------------------------------------------------------------------------
 We are interested in verifying the following invariants, which are 
 essential in proving the protocol specification.
------------------------------------------------------------------------------%

let J1 = new_definition (`J1`,
   "J1 = \s:^XState. (s `Rbit` = s `Sbit`) ==> (s `wire` = s `buffer`)") ;;

let J2 = new_definition (`J2`,
   "J2 = \s:^XState. (s `ack` = s `Sbit`) ==> (s `Sbit` = s `Rbit`)") ;;

let J = "J1 AND J2" ;;

%-----------------------------------------------------------------------------
  In order to prove above we still need the following assumption that states
  Next generates a diferent (abstract) bit.
-----------------------------------------------------------------------------%

let Distinct_Next = new_definition (`Distinct_Next`,
   "Distinct_Next Next = !n:num. (n = Next n) ==> F") ;;

let ALT_BIT_defs = map snd (definitions `ALT_BIT`) ;;

%-----------------------------------------------------------------------------
   First we verify that Init ==> J1 /\ J2
-----------------------------------------------------------------------------%

let AB_INV1 =
    prove("|== (Init IMP (J1 AND J2))",
    REWRITE_TAC ALT_BIT_defs
    THEN pred_DECOM_TAC 5 THEN XRULE_ASSUM_TAC SYM THEN RES_TAC) ;;


%-----------------------------------------------------------------------------
   Next, we verify that ALT_BIT invariant J1/\J2 is stable
-----------------------------------------------------------------------------%

let AB_INV2 = prove(
   "Distinct_Next Next ==> UNLESS (ALT_BIT Exp Next) (J1 AND J2) FF",
    REWRITE_TAC ALT_BIT_defs THEN DISCH_TAC
    THEN UPROG_UNFOLD_TAC
    THEN UNLESS_DECOM_TAC 3
    THEN pred_DECOM_TAC 4
    THEN EVERY_ASSUM (\thm. (REWRITE_TAC [SYM thm]) ?  ALL_TAC)) ;;

%-----------------------------------------------------------------------------

  Now we will prove the specification of ALT BIT:

      (0) J /\ wire=X  |-->  buffer=X

  to prove above first we prove:

      (1) J /\ ~Sbit=Rbit /\ wire=X UNLESS buffer=X
      (2) J /\  Sbit=Rbit /\ wire=X UNLESS buffer=X
      (3) J /\ ~Sbit=Rbit /\ wire=X ENSURES buffer=X
      (4) J /\  Sbit=Rbit /\ wire=X ENSURES buffer=X
      (5) J /\ ~Sbit=Rbit /\ wire=X |--> buffer=X
      (6) J /\  Sbit=Rbit /\ wire=X |--> buffer=X
      (7) J /\  wire=X |--> buffer=X

Below is the proofs of the UNLESS specs

-------------------------------------------------------------------------------%

let AB_SAFE1 = prove(
   "Distinct_Next Next ==>
    UNLESS (ALT_BIT Exp Next) 
           (J1 AND J2 AND (\s:^XState. ~(s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
           (\s. s `buffer` = X)",
    REWRITE_TAC ALT_BIT_defs THEN DISCH_TAC
    THEN UPROG_UNFOLD_TAC
    THEN UNLESS_DECOM_TAC 2
    THEN pred_DECOM_TAC 5 
    THEN XRULE_ASSUM_TAC SYM THEN RES_TAC) ;;


let AB_SAFE2 = prove(
   "Distinct_Next Next ==>
    UNLESS (ALT_BIT Exp Next) 
           (J1 AND J2 AND (\s:^XState. (s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
           (\s. s `buffer` = X)",
    REWRITE_TAC ALT_BIT_defs THEN DISCH_TAC
    THEN UPROG_UNFOLD_TAC
    THEN UNLESS_DECOM_TAC 2
    THEN pred_DECOM_TAC 6
    THEN XRULE_ASSUM_TAC SYM THEN RES_TAC
    THEN XRULE_ASSUM_TAC (SUBS [ASSUME "(s:^XState) `wire` = s `buffer`"])
    THEN XRULE_ASSUM_TAC SYM THEN RES_TAC ) ;;

%----------------------------------------------------------------------------
                   The proofs of ENSURES specs
----------------------------------------------------------------------------%

let AB_ENS1 =  prove(
   "RC [Send Exp Next; Receive; ACK] AB_Rd /\
    Distinct_Next Next 
    ==>
    ENSURES (ALT_BIT Exp Next)
            (J1 AND J2 AND (\s:^XState. ~(s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
            (\s. s `buffer` = X)",
    REWRITE_TAC [ALT_BIT] 
    THEN REPEAT STRIP_TAC    
    THEN ENSURES_DECOM_TAC 2 "Receive"
    THENL 
      [ IMP_RES_TAC (REWRITE_RULE [ALT_BIT] AB_UNITY) ;
        IMP_RES_TAC (REWRITE_RULE [ALT_BIT] AB_SAFE1) 
        THEN ASM_REWRITE_TAC[];
        UNDISCH_TAC "Distinct_Next Next" THEN REWRITE_TAC ALT_BIT_defs
        THEN ASSIGN_UNFOLD_TAC
        THEN pred_DECOM_TAC 8 ]) ;;

let AB_ENS2 = prove(
   "RC [Send Exp Next; Receive; ACK] AB_Rd /\
    Distinct_Next Next 
    ==>
    ENSURES (ALT_BIT Exp Next)
            (J1 AND J2 AND (\s:^XState. (s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
            (\s. s `buffer` = X)",
    REPEAT STRIP_TAC 
    THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
    THEN EXISTS_TAC 
         "(J1 AND J2 AND (\s:^XState. (s `Sbit` = s `Rbit`) /\ (s `wire` = X)))"
    THEN CONJ_TAC
    THENL
    [  MATCH_MP_TAC (QUAFF_RULE 3 ENSURES_REFL)
       THEN IMP_RES_TAC AB_UNITY ;
       REWRITE_TAC ALT_BIT_defs THEN pred_DECOM_TAC 2 
       THEN XRULE_ASSUM_TAC SYM
       THEN RES_TAC THEN ASM_REWRITE_TAC[]]) ;;

%--------------------------------------------------------------------------------
            the proofs of the locality of the predicates in the specs
--------------------------------------------------------------------------------%

let AB_LP1 = prove (
    "LocPred (ALT_BIT Exp Next)
             (J1 AND J2 AND
                (\s:^XState. ~(s `Sbit` = s `Rbit`) /\ (s `wire` = X)))",
    REWRITE_TAC ALT_BIT_defs THEN LocPred_DECOM_TAC) ;;

let AB_LP2 = prove (
    "LocPred (ALT_BIT Exp Next)
             (J1 AND J2 AND
                (\s:^XState. (s `Sbit` = s `Rbit`) /\ (s `wire` = X)))",
    REWRITE_TAC ALT_BIT_defs THEN LocPred_DECOM_TAC) ;;

let AB_LP3 = prove (
   "LocPred (ALT_BIT Exp Next) (\s:^XState. s `buffer` = X)",
    REWRITE_TAC ALT_BIT_defs THEN LocPred_DECOM_TAC) ;;


%-------------------------------------------------------------------------------
  Now the final proofs of ALT_BIT:
--------------------------------------------------------------------------------%

let AB_PROG1 = prove(
    "RC [Send Exp Next; Receive; ACK] AB_Rd /\
     Distinct_Next Next ==>
     LEADSTO (ALT_BIT Exp Next)
             (J1 AND J2 AND 
                 (\s:^XState. ~(s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
            (\s. s `buffer` = X)",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (QUAFF_RULE 3 LEADSTO_ENS_LIFT)
    THEN REWRITE_TAC [AB_LP1; AB_LP3]
    THEN IMP_RES_TAC AB_ENS1 THEN ASM_REWRITE_TAC[]) ;;

let AB_PROG2 = prove(
    "RC [Send Exp Next; Receive; ACK] AB_Rd /\
     Distinct_Next Next ==>
     LEADSTO (ALT_BIT Exp Next)
             (J1 AND J2 AND
                 (\s:^XState. (s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
            (\s. s `buffer` = X)",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (QUAFF_RULE 3 LEADSTO_ENS_LIFT)
    THEN REWRITE_TAC [AB_LP2; AB_LP3]
    THEN IMP_RES_TAC AB_ENS2 THEN ASM_REWRITE_TAC[]) ;;

let lemma = prove
    ("(J1 AND J2 AND (\s:^XState. s `wire` = X)) =
      ((J1 AND J2 AND 
                 (\s:^XState. ~(s `Sbit` = s `Rbit`) /\ (s `wire` = X)))
       OR
       (J1 AND J2 AND 
                 (\s:^XState. (s `Sbit` = s `Rbit`) /\ (s `wire` = X))))",
     pred_CYAN_TAC) ;;
     
let AB_PROG3 = prove(
    "RC [Send Exp Next; Receive; ACK] AB_Rd /\
     Distinct_Next Next ==>
     LEADSTO (ALT_BIT Exp Next)
             (J1 AND J2 AND (\s:^XState. s `wire` = X))
             (\s. s `buffer` = X)",
    SUBST1_TAC lemma THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC LEADSTO_SIMPLE_DISJ
    THEN IMP_RES_TAC AB_PROG1
    THEN IMP_RES_TAC AB_PROG2
    THEN ASM_REWRITE_TAC[]) ;;

close_theory() ;;
%---------------------------------------------------------------------------------

   End.
   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.ruu.nl

---------------------------------------------------------------------------------%
