%---------------------------------------------------------------------------------------
 This is the support file for the theory UPROG. This files contains some ML tools,
 such as tactics, to automate the verification of a UNITY program.
   
  Written by Wishnu Prasetya
  Dept. of Computer Science,
  Utrecht University
  Email : wishnu@cs.ruu.nl
---------------------------------------------------------------------------------------%

%--------------------------------------------------------------------------------------
                    some standard ML abbreviations
---------------------------------------------------------------------------------------%

let State = ":*var -> *value" ;;
let Pred = ":^State -> bool" ;;
let Expr = ":^State -> *value" ;;
let Action = ":^State -> ^State -> bool" ;;
let Actions = ":^Action -> bool" ;;
let Tran = ":^State -> ^State" ;;
let Trans = ":^Tran -> bool" ;;
let Vars = ":*var -> bool" ;;
let Uprog = ":^Actions # ^Pred # ^Vars # ^Vars" ;;

%------------ tactic to help proving that a program is a UNITY program -----------

  For example

  let A1 = "ASGL [(x:*var,E1:^Expr); (y,E2)]" ;;
  let A2 = "CASE [(b:^Pred,^A1); (NOT b, (x ASG E3))]" ;;
  let A3 = "CASE [(c:^Pred,^A1); (NOT c,^A2); (NOT b,^A1) ]" ;;
  let Rd = "[x:*var;y;z;b;c]" ;;
  let Wr = "[x:*var; y]" ;;

  let Pr = "UPROG (^Rd) (^Wr) (In:^Pred) [^A1;^A2; ^A3]" ;;


        "RC [^A1;^A2;^A3]  ^Rd ==> UNITY ^Pr"
       ----------------------------------------  UNITY_DECOM_TAC 5
                       proven

----------------------------------------------------------------------------------%

let UNITY_DECOM_TAC n  =
    STRIP_TAC
    THEN MATCH_MP_TAC UPROG_UNITY
    THEN ASM_REWRITE_TAC[]
    THEN REPEAT CONJ_TAC
    THEN REWRITE_TAC [EVERY_DEF; MEMBER]
    THEN BETA_TAC 
    THENL [REWRITE_TAC [NOT_CONS_NIL] ; 
           ALL_TAC ;
           REWRITE_TAC []] 
    THEN (REPEAT_FIN n
          (BETA_TAC
           THEN REPEAT CONJ_TAC
           THEN 
            (MATCH_MP_TAC ASSIGN_SAT_WC
             ORELSE MATCH_MP_TAC ASSIGN2_SAT_WC
             ORELSE MATCH_MP_TAC ASGL_SAT_WC
             ORELSE MATCH_MP_TAC THEN_SAT_WC
             ORELSE MATCH_MP_TAC CASE_SAT_WC)
           THEN REWRITE_TAC [MAP; FST; SND; EVERY_DEF; L2P_DEF; MEMBER])) ;;

%----------------- a Rule & Tactic to handle ---------------------------------------
 the recusive unfolding of the definition of multiple assignment.
 for example 
   
    let thm = ASSUME "f = ASGL [(x:*var,E1:^Expr) ; (y,E2) ; (z,E3) ] s" ;;

 then applying ASGL_UNFOLD_RULE to above yields:

 . |- f = (\v.
          ((v = x) => E1 s | ((v = y) => E2 s | ((v = z) => E3 s | s v))))

 the tactic works analogously.
-----------------------------------------------------------------------------------%

let ASSIGN_UNFOLD_RULE =
    BETA_RULE o REWRITE_RULE 
       [ASSIGN_DEF; ASSIGN2_DEF; 
        ASGL_DEF; MAP; ASGL_DEF; EVERY_DEF; update_DEF] ;;

let ASSIGN_UNFOLD_TAC =
    REWRITE_TAC 
       [ASSIGN_DEF; ASSIGN2_DEF; 
        ASGL_DEF; MAP; ASGL_DEF; EVERY_DEF; update_DEF] 
    THEN BETA_TAC ;;

%----------------- a Rule & Tactic to handle ---------------------------------------
 the recusive unfolding of the definition of CASE statement.
 For example 
   
    let thm = ASSUME "f = CASE [(b1:^Pred,A1:^Tran) ; (b2,A2) ; (b3,A3) ] s" ;;



 then applying CASE_UNFOLD_RULE to above yields:

    . |- f = (b1 s => A1 s | (b2 s => A2 s | (b3 s => A3 s | s)))

 the tactic works analogously.
-----------------------------------------------------------------------------------%

let COND_UNFOLD_RULE =
    BETA_RULE o REWRITE_RULE 
      [THEN_DEF; MAP; CASE_DEF; EVERY_DEF; I_THM] ;;

let COND_UNFOLD_TAC =
    REWRITE_TAC [THEN_DEF; MAP; CASE_DEF; EVERY_DEF; I_THM]
    THEN BETA_TAC ;;




%---------------------------------------------------------------------------------
 If the expressions involved in |== P are only simple arithmetics then 
 pred_DECOM_TAC may solve it
---------------------------------------------------------------------------------%

let pred_DECOM_TAC n =
    REWRITE_TAC (map snd (definitions `predicate`))
    THEN REPEAT_FIN n
      (BETA_TAC
       THEN CONV_TAC (DEPTH_CONV num_EQ_CONV)
       THEN CONV_TAC (DEPTH_CONV string_EQ_CONV)
       THEN REWRITE_TAC[]
       THEN (STRIP_TAC ORELSE ALL_TAC) 
       THEN (CONJ_TAC ORELSE ALL_TAC)
       THEN (COND_CASES_TAC ORELSE ALL_TAC)
       THEN IMP_RES_TAC F_IMP
       THEN RES_TAC
       THEN ASM_REWRITE_TAC[]) ;;

%-----------------------------------------------------------------------------   
       UPROG_UNFOLD_TAC : change list representation of a program to a 
                          predicate representation.

       UNLESS_DECOM_TAC : decompose an UNLESS expression down to a
                          predicate level
-------------------------------------------------------------------------------%

let UPROG_UNFOLD_TAC =
    REWRITE_TAC [UPROG_DEF; MAP] ;;

let UNLESS_DECOM_TAC n =
    REWRITE_TAC[UNLESS; PROG] 
    THEN RESTRICT_DEF_TAC
    THEN REWRITE_TAC [L2P_DEF; MEMBER]
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC SYM
    THEN ASM_REWRITE_TAC[] 
    THEN REWRITE_TAC [HOA_DEF; F2R_DEF]
    THEN REPEAT_FIN n
         (ASSIGN_UNFOLD_TAC THEN COND_UNFOLD_TAC) ;;

%-----------------------------------------------------------------------------
     ENSURES_DECOM_TAC : decompose an ENSURES expression down to a
                         predicate level
-------------------------------------------------------------------------------%

let ENSURES_DECOM_TAC n FAct =
    let RAct = "F2R ^FAct" in
    REWRITE_TAC [ENSURES]
    THEN REPEAT STRIP_TAC
    THENL 
     [ALL_TAC ;
      ALL_TAC ;
      RESTRICT_DEF_TAC
      THEN EXISTS_TAC RAct %----- select F2R FAct -------------%
      THEN CONJ_TAC
      THENL [ %--- case1 : prove that FAct is in the program ---%
              REWRITE_TAC [UPROG_DEF; PROG; MAP; L2P_DEF; MEMBER] ;
              %--- case2 : simplify the Hoare triple -----------%
             REWRITE_TAC [L2P_DEF; MEMBER; PROG]
             THEN REPEAT STRIP_TAC
             THEN XRULE_ASSUM_TAC SYM
             THEN ASM_REWRITE_TAC[] 
             THEN REWRITE_TAC [HOA_DEF; F2R_DEF]
             THEN REPEAT_FIN n
                 (ASSIGN_UNFOLD_TAC THEN COND_UNFOLD_TAC)]] ;;

%-----------------------------------------------------------------------------
   Here is a tactic to verify locality of a predicate.
-------------------------------------------------------------------------------%

letrec SPEC_ONE_BY_ONE_TAC VarList =
     if VarList=[] then ALL_TAC
     else SPEC_ONE_BY_ONE_TAC (tl VarList)
          THEN FIRST_ASSUM 
               (\thm. ASSUME_TAC (SPEC (hd VarList) thm)) ;;

let LocPred_DECOM_TAC (asml,g)  =
    let prog  = (rand o rator) g in
    let Vlist = (rand o rator o rator o rator) prog in
    let vlist = fst (dest_list Vlist) in
    (REWRITE_TAC [LocPred; UPROG_DEF; READ] 
     THEN RESTRICT_DEF_TAC
     THEN REWRITE_TAC [L2P_DEF; MEMBER]  
     THEN REWRITE_TAC (map snd (definitions `predicate`))
     THEN BETA_TAC
     THEN REPEAT STRIP_TAC 
     THEN SPEC_ONE_BY_ONE_TAC vlist
     THEN XRULE_ASSUM_TAC (REWRITE_RULE[])
     THEN ASM_REWRITE_TAC[]) (asml,g) ;;

%--------------------------------------------------------------------------------

   End.
   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.ruu.nl

---------------------------------------------------------------------------------%





