%-----------------------------------------------------------------------------------

  Written by Wishnu Prasetya
  Dept. of Computer Science,
  Utrecht University
  Email : wishnu@cs.ruu.nl

  The theory UPROG below defines the semantic of UNITY programming language. 
  Our version of UNITY's syntax is as follows:


     <UNITY program>   ::= UPROG <Read Vars> <Write Vars> <Init> <Trans>>
     <Read Vars>       ::= <Var list>
     <Write Vars>      ::= <Var list>
     <Init>            ::= <Pred>
     <Trans>           ::=  <Transition list>

     <Transition>      ::= <Assignment> | <Conditional>

     <Assignment>      ::= <Simple Asg> | <Simultant Asg> | <General Asg>
     <Simple Asg>      ::= <Var> ASG <Expr>
     <Simultant Asg>   ::= (<Var>,<Var>) ASG2 (<Expr>,<Expr>)
     <General Asg>     ::= ASGL <(Var,Expr) list>
     
     <Conditional>     ::= <THEN transition> | <CASE transition>
     <THEN transition> ::= <Pred> THEN <Transition>
     <CASE transition> ::= CASE <(Pred,Transition) list>

  A transition is a functional action, hence it is deterministic.

----------------------------------------------------------------------------------%

load_library `WF` ;;
load_theory `def_UNITY` ;;
new_theory `UPROG` ;;
load_library `string`;;
loadf `my_misc`;;

let Autoload thy =    
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `variables` ;;
Autoload `def_UNITY` ;;
%----------------------- Types definition ------------------------------------------%

let State = ":*var -> *value" ;;
let Pred = ":(*var->*value)->bool" ;;
let Action = ":(*var->*value)->(*var->*value)->bool" ;;
let ACTIONS = ":((*var->*value)->(*var->*value)->bool)->bool" ;;
let VARS = ":*var->bool" ;;
let Uprog = ":^ACTIONS # ^Pred # ^VARS # ^VARS" ;;

let Expr = ":^State -> *value" ;;
let Tran = ":^State -> ^State" ;;
let Trans = ":^Tran -> bool" ;;

%----------------------- Semantic of Assignments -----------------------------------
 
In case of an simultant assignment, it is simultant in the sense that the 
expressions to be assigned are all evaluated in the begin state, but the order
of the variables determines the order of the assignments, so:

    {x=X /\ y=Y} x,y := y+1,y+1 {x=Y+1 /\ y=Y+1}

but:

    {x=X} x,x := 2,3 {x=2}

-----------------------------------------------------------------------------------%

let update_DEF = new_definition (`update_DEF`,
    "update x val (s:^State) = (\v. (v=x) => val |  s v)") ;;

let ASSIGN_DEF = new_infix_definition (`ASSIGN_DEF`,
    "ASG x Expr (s:^State) = update x (Expr s) s") ;;
     
let ASSIGN2_DEF =  new_infix_definition (`ASSIGN2_DEF`,
    "ASG2 (x1,x2) (E1,E2) (s:^State) = update x1 (E1 s) (update x2 (E2 s) s)") ;;

let ASGL_DEF = new_list_rec_definition (`ASGL_DEF`,
   "(ASGL [] (s:^State) = s) /\ 
    (!Asg AsgLst. ASGL (CONS Asg AsgLst) s =
     update (FST Asg) (SND Asg s) (ASGL AsgLst s))") ;;

%----------------------- Semantic of Conditionals ------------------------------------%

let THEN_DEF = new_infix_definition (`THEN_DEF`,
    "THEN b (Tr:^Tran) s = (b s => Tr s | s)") ;;

let CASE_DEF = new_list_rec_definition (`CASE_DEF`,
   "(CASE [] (s:^State) = s) /\
    (!Ctr CtrLst. CASE (CONS Ctr CtrLst) s =
            (FST Ctr s => SND Ctr s | CASE CtrLst s))") ;;


%------------------------ Semantic of UPROG  -----------------------------------------%

let L2P_DEF = new_definition (`L2P_DEF`, "L2P (s:* list) x = IS_EL x s") ;;

let F2R_DEF = new_definition (`F2R_DEF`, "F2R f (x:*) (y:**) = (y = (f x))") ;;

let UPROG_DEF = new_definition (`UPROG_DEF`,
    "UPROG (RvL:(*var)list) (WvL:(*var)list) (In:^Pred) (TrL:(^Tran) list) = 
           ((L2P (MAP F2R TrL)), In, L2P RvL, L2P WvL)") ;;


%---------------------- Always Enabledness of transitions ----------------------------
  
  ALL functional actions, or transitions, are always enabled.

-------------------------------------------------------------------------------------%

let ALWAYS_EN_TRANS = prove_thm (`ALWAYS_EN_TRANS`,
    "!Tr:^Tran. ALWAYS_ENABLED (F2R Tr)",
    REWRITE_TAC [ALWAYS_ENABLED; F2R_DEF]
    THEN REPEAT GEN_TAC
    THEN EXISTS_TAC "(Tr:^Tran) s"
    THEN REWRITE_TAC[]) ;;

%--------------------- Write constraint over the transitions -------------------------

 The theorems below states the sufficient conditions under which a variable set
 is ignored by a transition.

 This gives a way to recursively verify whether a given UNITY program satisfies
 its write constraint.

-------------------------------------------------------------------------------------%

let ASSIGN_CASE_TAC x y =
    ASM_CASES_TAC "^x=^y"
    THEN ASM_REWRITE_TAC[]
    THEN CONV_TAC (DEPTH_CONV COND_CONV)
    THEN RULE_ASSUM_TAC (\thm. if not (is_eq(concl thm) )
                                  then REWRITE_RULE [ASSUME "^x=^y"] thm
                                  else thm)
    THEN RES_TAC ;;

let ASSIGN_SAT_WC = prove_thm (`ASSIGN_SAT_WC`, 
    "!W x (E:^Expr). W x ==> (NOT W) IG_BY (F2R (x ASG E))",
     REWRITE_TAC [pNOT_DEF; IG_BY_DEF; ASSIGN_DEF; update_DEF; F2R_DEF;
                  Pj_EQ]
     THEN BETA_TAC THEN REPEAT STRIP_TAC 
     THEN ASM_REWRITE_TAC [] THEN BETA_TAC
     THEN ASSIGN_CASE_TAC "x':*var" "x:*var") ;;

let ASSIGN2_SAT_WC = prove_thm (`ASSIGN2_SAT_WC`, 
    "!W x1 x2 E1 (E2:^Expr). 
        W x1 /\ W x2 ==> (NOT W) IG_BY (F2R ((x1,x2) ASG2 (E1,E2)))",
     REWRITE_TAC [pNOT_DEF; IG_BY_DEF; ASSIGN2_DEF; update_DEF; F2R_DEF;
                  Pj_EQ]
     THEN BETA_TAC THEN REPEAT STRIP_TAC 
     THEN ASM_REWRITE_TAC [] THEN BETA_TAC
     THEN ASM_CASES_TAC "(x:*var)=x1"
     THEN ASM_REWRITE_TAC[]
     THEN CONV_TAC (DEPTH_CONV COND_CONV)
     THENL 
     [ ASSIGN_CASE_TAC "x:*var" "x1:*var" ;
       ASSIGN_CASE_TAC "x:*var" "x2:*var" ]) ;;


let ASGL_SAT_WC = prove_thm (`ASGL_SAT_WC`,
    "!AsgL. 
     ALL_EL W (MAP FST AsgL) ==> (NOT W) IG_BY (F2R ((ASGL AsgL):^Tran))",
     LIST_INDUCT_TAC
     THEN UNDISCH_ALL_TAC
     THEN REWRITE_TAC [IG_BY_DEF; ASGL_DEF; F2R_DEF; MAP; ALL_EL]
     THEN REPEAT STRIP_TAC 
     THEN RES_TAC THEN ASM_REWRITE_TAC[] 
     THEN SUBGOAL_THEN 
          "((s:^State) Pj (NOT W)) = ((ASGL AsgL s) Pj (NOT W))"
          SUBST1_TAC
     THENL
     [ FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC ;
       MATCH_MP_TAC (QUAFF_RULE 3 
                    (BETA_RULE 
                    (REWRITE_RULE 
                     [IG_BY_DEF; ASSIGN_DEF; F2R_DEF] ASSIGN_SAT_WC)))
       THEN EXISTS_TAC "FST (h:*var#^Expr)"
       THEN EXISTS_TAC "\t:^State. SND (h:*var#^Expr) s"
       THEN BETA_TAC THEN ASM_REWRITE_TAC[]
      ]) ;;

         
let THEN_SAT_WC = prove_thm (`THEN_SAT_WC`,
    "!W G (Tr:^Tran). 
        (NOT W) IG_BY (F2R Tr) ==> (NOT W) IG_BY (F2R (G THEN Tr))",
     REWRITE_TAC [IG_BY_DEF; F2R_DEF; THEN_DEF; Pj_EQ]
     THEN BETA_TAC
     THEN REPEAT STRIP_TAC
     THEN ASM_CASES_TAC "(G:^Pred) s" 
     THEN RES_TAC THEN ASM_REWRITE_TAC[]
     THEN FIRST_ASSUM (MATCH_MP_TAC o (QUAFF_RULE 3))
     THEN ASM_REWRITE_TAC[]) ;;

let CASE_SAT_WC = prove_thm (`CASE_SAT_WC`,
    "!CtrLst. 
        ALL_EL (\Tr. (NOT W) IG_BY (F2R Tr)) (MAP SND CtrLst) 
        ==> 
       (NOT W) IG_BY (F2R ((CASE CtrLst):^Tran))",
    LIST_INDUCT_TAC 
    THEN UNDISCH_ALL_TAC 
    THEN REWRITE_TAC [CASE_DEF; F2R_DEF; IG_BY_DEF; MAP; ALL_EL]
    THEN BETA_TAC
    THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]
    THEN ASM_CASES_TAC "FST (h:(^Pred) # (^Tran)) s" 
    THEN ASM_REWRITE_TAC []
    THEN CONV_TAC (DEPTH_CONV COND_CONV)
    THENL 
    [ XRULE_ASSUM_TAC (SPECL ["s:^State"; "SND (h:^Pred#^Tran) s"]) ;
      XRULE_ASSUM_TAC (SPECL ["CASE CtrLst (s:^State)"; "s:^State"])]
    THEN XRULE_ASSUM_TAC (REWRITE_RULE[])
    THEN ASM_REWRITE_TAC[]) ;;

%--------------------- Read constraint over the transitions -------------------------

  Automatic checking of Read Constraint is NOT implemented.

  If all variables appearing at the right hand side of the assignments and in the
  guard of the conditionals of a program Pr form a subset of READ Pr, then one can 
  prove that the Read Constraint is satisfied. But since we do not define an
  expression laguage for UNITY in HOL we cannot do this verification. It is then
  the user's responsibility to guarantee that his program meets the Read Constraint.

  To denote this obligation we define the constant RC below. 
  RC TrL RvL means we still have to verify that every variable in RvL is invicible
  to every transition in TrL.

-------------------------------------------------------------------------------------%

let RC_DEF = new_definition 
  (`RC_DEF`,
   "RC (TrL:^Tran list) (RvL:*var list) = 
    ALL_EL (\Tr. (NOT(L2P RvL)) INVI (F2R Tr)) TrL") ;;

%------------------------------------------------------------------------------
 Using the previous theorems about always enabledness and write constraint 
 satisfiability, the theorem UPROG_UNITY below gives a condition sufficent to 
 conclude when an UPROG (syntactical) construct really represents a UNITY 
 program.
------------------------------------------------------------------------------%

let ALL_EL_IS_EL = prove
  ("!s:(*)list. ALL_EL P s = (!x. IS_EL x s ==> P x)",
    LIST_INDUCT_TAC
    THEN REWRITE_TAC [ALL_EL; IS_EL]
    THEN ASM_REWRITE_TAC[]
    THEN GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC 
    THENL 
    [ ASM_REWRITE_TAC[] ;
      FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
      XRULE_ASSUM_TAC (REWRITE_RULE[] o (SPEC "h:*"))
      THEN ASM_REWRITE_TAC[] ;
      RES_TAC ]) ;;

let ALL_EL_MAP = prove
  ("!l:(*)list. ALL_EL P (MAP (f:*->**) l) = ALL_EL (P o f) l",
    LIST_INDUCT_TAC
    THEN REWRITE_TAC [ALL_EL; MAP]
    THEN POP_ASSUM (\thm. REWRITE_TAC [BETA_RULE thm])
    THEN REWRITE_TAC [o_DEF]
    THEN BETA_TAC THEN REWRITE_TAC []) ;;

let IS_EL_MAP = prove
   ("!l:(*)list. IS_EL (x:**) (MAP f l) = ?y. (x = f y) /\ IS_EL y l",
    LIST_INDUCT_TAC
    THEN REWRITE_TAC [IS_EL; MAP]
    THEN ASM_REWRITE_TAC[]
    THEN GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
    THENL 
      [EXISTS_TAC "h:*" THEN ASM_REWRITE_TAC[] ;
       EXISTS_TAC "y:*" THEN ASM_REWRITE_TAC[] ;
       ASM_REWRITE_TAC[] ;
       DISJ2_TAC
       THEN EXISTS_TAC "y:*" THEN ASM_REWRITE_TAC[]]) ;;


let UPROG_UNITY = prove_thm(`UPROG_UNITY`,
    "!(RvL:(*var)list) (WvL:(*var)list) (In:^Pred) (TrL:(^Tran) list).
        ~(TrL=[])  /\
        ALL_EL (\Tr. (NOT (L2P WvL)) IG_BY (F2R Tr)) TrL /\
        RC TrL RvL /\
        ALL_EL (\x. IS_EL x RvL) WvL
        ==>
        UNITY (UPROG RvL WvL In TrL)",
    REWRITE_TAC [UPROG_DEF; UNITY; RC_DEF]
    THEN RESTRICT_DEF_TAC
    THEN REPEAT GEN_TAC 
    THEN DISJ_CASES_TAC (ISPEC "TrL:^Tran list" list_CASES)
    THEN UNDISCH_ALL_TAC THEN STRIP_TAC
    THEN ASM_REWRITE_TAC[]
    THEN REWRITE_TAC [ALL_EL_IS_EL; L2P_DEF; IS_EL_MAP]
    THEN BETA_TAC THEN REWRITE_TAC [IS_EL]
    THEN REPEAT STRIP_TAC
    THENL 
    [ %-- case 1: non-emptyness --%
      CONV_TAC SWAP_EXISTS_CONV
      THEN EXISTS_TAC "h:^Tran" 
      THEN EXISTS_TAC "F2R (h:^Tran)" THEN REWRITE_TAC[] ;
      %-- case 2: always enabledness --%
      ASM_REWRITE_TAC [ALWAYS_EN_TRANS] ;
      ASM_REWRITE_TAC [ALWAYS_EN_TRANS] ;
      %-- case3: write constraint --%
      ASM_REWRITE_TAC[] 
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[] ;
      ASM_REWRITE_TAC[] 
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[] ;
      %-- case4: inclusion of write in read --%
      FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
      %-- case5: read constraint --%
      ASM_REWRITE_TAC[] 
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[] ;
      ASM_REWRITE_TAC[] 
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[] ]) ;;

close_theory();;

%---------------------------------------------------------------------------------

   End.
   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.ruu.nl

---------------------------------------------------------------------------------%











