%---------------------------------------------------------------------------------
   This file contains a theory of a progress operator called REACH. Roughly 
   speaking, REACH Pr J p q means that the program Pr will progress from J/\p to
   q and further more J is stable.

   p and q are restricted to specify no other vraible than the read variables 
   of Pr. J can however specify any variable. Typically, it specifies the state
   of the input variables of Pr.

   The theory is based on UNITY programming logic.
   

   Author:

     Wishnu Prasetya
     Dept. of Computer Science
     Utrecht University
     the Netherlands

     email: wishnu@cs.ruu.nl
---------------------------------------------------------------------------------% 

loadf `UU_UNITY` ;;
new_theory `reach` ;;
loadf `OLD_RES`;;

%---------------------------------------------------------------------------------
   the input variables of a program is the non-writeable part of its read
   variables.
---------------------------------------------------------------------------------%

let INP = new_definition 
  (`INP`,
   "INP (Pr:^Uprog) = (NOT (WRITE Pr)) AND (READ Pr)") ;;

%---------------------------------------------------------------------------------
     if a predicate confined by the write variables of a program then it is 
     a local variable of that process.
---------------------------------------------------------------------------------%

let CONF_WRITE_Loc = prove_thm
  (`CONF_WRITE_Loc`,
   "!(Pr:^Uprog) p. UNITY Pr /\ CONF (WRITE Pr) p ==> LocPred Pr p",
    ONCE_REWRITE_TAC [ID_UPROG]
    THEN REWRITE_TAC [UNITY; LocPred_DEF; PROG; INIT; READ; WRITE]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CONF_MONO
    THEN EXISTS_TAC "WRITE (Pr:^Uprog)"
    THEN ASM_REWRITE_TAC [pIMP_DEF; pSEQ_DEF]) ;;
    
%---------------------------------------------------------------------------------
  Now we define REACH as, roughly speaking, the least transitive and left 
  disjunctive closure of ENSURES, conditioned by some stable predicate.
---------------------------------------------------------------------------------%
 
let B_ENS = new_definition 
  (`B_ENS`,
   "B_ENS (Pr:^Uprog) J p q =
      ENSURES Pr (p AND J) q /\ 
      STABLE Pr J /\ 
      CONF (WRITE Pr) p /\ CONF (WRITE Pr) q") ;;

let REACH = new_definition 
  (`REACH`,
   "REACH (Pr:^Uprog) J = TDC (B_ENS Pr J)") ;;

%---------------------------------------------------------------------------------
  Here are the most basic properties of REACH:
     (1) REACH includes B_ENS
     (2) REACH transitive
     (3) left disjunctive
---------------------------------------------------------------------------------%

let REACH_ENS_LIFT = prove_thm
  (`REACH_ENS_LIFT`,
   "!(Pr:^Uprog) J. (B_ENS Pr J) SUBREL (REACH Pr J)",
   REWRITE_TAC [REACH; TDC_LIFT]) ;;

let REACH_ENS_LIFT_TAC =
    let thm = REWRITE_RULE [SUBREL; B_ENS] REACH_ENS_LIFT in
    MATCH_MP_TAC (QUAFF_RULE 3 thm)
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_TRANS = prove_thm
  (`REACH_TRANS`,
   "!(Pr:^Uprog) J. TRANS (REACH Pr J)",
   REWRITE_TAC [TDC_TRANS; REACH]) ;;

let REACH_TRANS_TAC trm =
    MATCH_MP_TAC (REWRITE_RULE [TRANS_DEF] REACH_TRANS)
    THEN EXISTS_TAC trm 
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_LDISJ = prove_thm
  (`REACH_LDISJ`,
   "!(Pr:^Uprog) J. LDISJ (REACH Pr J)",
   REWRITE_TAC [REACH; TDC_LDISJ] ) ;;

let REACH_LDISJ_TAC =
    MATCH_MP_TAC (REWRITE_RULE [LDISJ_DEF] REACH_LDISJ)
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

%---------------------------------------------------------------------------------
                    Induction theorems for REACH
---------------------------------------------------------------------------------%

let MK_2ABS_CONV x y f =
    SYM ((DEPTH_CONV BETA_CONV) "(\^x ^y. ^f) ^x ^y") ;;

let REACH_INDUCT1 = prove_thm
  (`REACH_INDUCT1`,
   "!X (Pr:^Uprog) J. 
       (B_ENS Pr J SUBREL X) /\ TRANS X /\ LDISJ X
       ==> 
       (!p q. REACH Pr J p q ==> X p q)",
    REPEAT GEN_TAC THEN
    REWRITE_TAC [REACH; SUBREL; REWRITE_RULE [SUBREL] TDC_INDUCT1]) ;;

let REACH_INDUCT1_TAC = 
    REPEAT GEN_TAC
    THEN MATCH_MP_TAC REACH_INDUCT1
    THEN REWRITE_TAC [SUBREL; B_ENS; TRANS_DEF; LDISJ_DEF]
    THEN BETA_TAC 
    THEN REPEAT STRIP_TAC ;;

let REACH_INDUCT3 = prove_thm 
  (`REACH_INDUCT3`,
   "!X (Pr:^Uprog) H.
       (!p q. B_ENS Pr H p q ==> X p q) 
       /\
       (!p q r. REACH Pr H p q /\ X p q /\ REACH Pr H q r /\ X q r
                  ==> X p r) 
       /\
       (!W q. (?p. W p) /\ (!p::W. REACH Pr H p q) /\ (!p::W. X p q)
                ==> X (??p::W.p) q) 
       ==> 
       (!p q. REACH Pr H p q ==> X p q)",
    REPEAT_FIN 4 STRIP_TAC
    THEN REWRITE_TAC [REACH]
    THEN MATCH_MP_TAC (REWRITE_RULE [SUBREL] TDC_INDUCT2)
    THEN REWRITE_TAC [SYM(SPEC_ALL REACH); rINTER; TRANS_DEF; LDISJ_DEF]
    THEN BETA_TAC
    THEN REPEAT STRIP_TAC
    THENL
    [ IMP_RES_TAC (REWRITE_RULE [SUBREL] REACH_ENS_LIFT) ;
      RES_TAC ;
      IMP_RES_TAC (REWRITE_RULE [TRANS_DEF] REACH_TRANS) ;
      RES_TAC ;
      REACH_LDISJ_TAC
      THENL [ EXISTS_TAC "x:^Pred" THEN ASM_REWRITE_TAC[] ; ALL_TAC]
      THEN UNDISCH_ALL_TAC THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
      THEN RES_TAC ;
      FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_ALL_TAC THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
      THENL [ EXISTS_TAC "x:^Pred" THEN ASM_REWRITE_TAC[] ;
              RES_TAC ; RES_TAC ]]) ;;

let REACH_INDUCT3_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC REACH_INDUCT3
   THEN BETA_TAC
   THEN REPEAT STRIP_TAC ;;

%---------------------------------------------------------------------------------
                   REACH property can only valid in a UNITY program
---------------------------------------------------------------------------------%

let REACH_IMP_UNITY = prove_thm
  (`REACH_IMP_UNITY`,
   "!(Pr:^Uprog) J p q. REACH Pr J p q ==> UNITY Pr",
    REWRITE_TAC [REACH; TDC_DEF]
    THEN REPEAT STRIP_TAC
    THEN CONV_TAC (MK_2ABS_CONV "p:^Pred" "q:^Pred")
    THEN POP_ASSUM MATCH_MP_TAC
    THEN REWRITE_TAC [SUBREL; TRANS_DEF; LDISJ_DEF; 
                      B_ENS; ENSURES]
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;;

%---------------------------------------------------------------------------------
                REACH implies confinement by write variables
---------------------------------------------------------------------------------%

let REACH_IMP_CONF = prove_thm
  (`REACH_IMP_CONF`,
   "!(Pr:^Uprog) J p q. 
         REACH Pr J p q ==> CONF (WRITE Pr) p /\ CONF (WRITE Pr) q",
    REPEAT GEN_TAC
    THEN CONV_TAC (RAND_CONV (MK_2ABS_CONV "p:^Pred" "q:^Pred"))
    THEN REACH_INDUCT1_TAC
    THEN CONF_TAC 3
    THENL
    [ EXISTS_TAC "x:^Pred" THEN ASM_REWRITE_TAC[] ;
      UNDISCH_ALL_TAC THEN  RESTRICT_DEF_TAC 
      THEN REPEAT STRIP_TAC THEN RES_TAC ;
      UNDISCH_ALL_TAC THEN  RESTRICT_DEF_TAC 
      THEN REPEAT STRIP_TAC THEN RES_TAC ]) ;;

%---------------------------------------------------------------------------------
                         REACH implies stability
---------------------------------------------------------------------------------%

let REACH_IMP_STABLE = prove_thm
  (`REACH_IMP_STABLE`,
   "!(Pr:^Uprog) J p q. REACH Pr J p q ==> STABLE Pr J",
    REPEAT GEN_TAC
    THEN CONV_TAC (RAND_CONV (MK_2ABS_CONV "p:^Pred" "q:^Pred"))
    THEN REACH_INDUCT1_TAC
    THEN XRULE_ASSUM_TAC RESTRICT_SHIFT
    THEN RES_TAC) ;;

%---------------------------------------------------------------------------------
                         REACH implies progress
---------------------------------------------------------------------------------%

let REACH_IMP_LEADSTO = prove_thm
  (`REACH_IMP_LEADSTO`,
   "!(Pr:^Uprog) J p q. 
           (WRITE Pr) CONF J /\ REACH Pr J p q ==> LEADSTO Pr (p AND J) q",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC REACH_IMP_UNITY
    THEN CONV_TAC (MK_2ABS_CONV "p:^Pred" "q:^Pred")
    THEN UNDISCH_TAC "REACH (Pr:^Uprog) J p q"
    THEN REACH_INDUCT3_TAC
    THENL
    [ %--- ensures ---%
      XRULE_ASSUM_TAC (REWRITE_RULE [B_ENS])
      THEN MATCH_MP_TAC (QUAFF_RULE 3 LEADSTO_ENS_LIFT)
      THEN ASM_REWRITE_TAC[]
      THEN CONJ_TAC THEN MATCH_MP_TAC CONF_WRITE_Loc
      THEN ASM_REWRITE_TAC[] THEN CONF_TAC 2 ;
      %--- transitivity ---%
      MATCH_MP_TAC LEADSTO_TRANS
      THEN EXISTS_TAC "q AND (J:^Pred)" THEN ASM_REWRITE_TAC[]
      THEN (SUBST1_TAC o SYM o ISPEC "q AND (J:^Pred)") pOR_UNIT
      THEN ONCE_REWRITE_TAC [pOR_SYM]
      THEN SUBST1_TAC (prove ("(p AND J) = ((p AND J) AND (J:^Pred))",
                       pred_CYAN_TAC))
      THEN MATCH_MP_TAC (QUAFF_RULE 3 LEADSTO_PSP)
      THEN IMP_RES_TAC REACH_IMP_STABLE
      THEN ASM_REWRITE_TAC [SYM(SPEC_ALL STABLE)]
      THEN CONJ_TAC THEN MATCH_MP_TAC CONF_WRITE_Loc
      THEN ASM_REWRITE_TAC[] THEN CONF_TAC 2 ;
      %--- disjunctivity ---%
      REWRITE_TAC [ BETA_RULE (ISPECL ["W:^Pred->bool"; "\p:^Pred. p"] 
                    qAND_RIGHT_DISTR_OR) ]
      THEN MATCH_MP_TAC
           (GEN_ALL (BETA_RULE (ISPECL 
             ["Pr:^Uprog"; "W:^Pred->bool"; "\p:^Pred. p AND J"] 
             LEADSTO_DISJ_LEFT)))
      THEN ASM_REWRITE_TAC [Non_FF_lemma]
      THEN EXISTS_TAC "p:^Pred" THEN ASM_REWRITE_TAC []
      ] ) ;;

%---------------------------------------------------------------------------------
                           Monotonicity of REACH
---------------------------------------------------------------------------------%

let REACH_IMP_LIFT = prove_thm
  (`REACH_IMP_LIFT`,
   "!(Pr:^Uprog) J p q.
         UNITY Pr /\ STABLE Pr J /\ 
         (WRITE Pr) CONF p /\ (WRITE Pr) CONF q /\
         |== (p IMP q) 
         ==> 
         REACH Pr J p q",
    REPEAT STRIP_TAC
    THEN REACH_ENS_LIFT_TAC
    THEN ASM_REWRITE_TAC[]
    THEN MATCH_MP_TAC (QUAFF_RULE 3 ENSURES_IMP_LIFT)
    THEN ASM_REWRITE_TAC[]
    THEN UNDISCH_TAC "|== (p IMP (q:^Pred))"
    THEN pred_JADE_TAC 3) ;;

let REACH_IMP_gLIFT = prove_thm
  (`REACH_IMP_gLIFT`,
   "!(Pr:^Uprog) J p q.
         UNITY Pr /\ STABLE Pr J /\ 
         (WRITE Pr) CONF p /\ (WRITE Pr) CONF q /\
         |== ((p AND J) IMP q) 
         ==> 
         REACH Pr J p q",
    REPEAT STRIP_TAC
    THEN REACH_ENS_LIFT_TAC
    THEN ASM_REWRITE_TAC[]
    THEN MATCH_MP_TAC (QUAFF_RULE 3 ENSURES_IMP_LIFT)
    THEN ASM_REWRITE_TAC[]) ;;

let REACH_MONO = prove_thm
  (`REACH_MONO`,
  "!(Pr:^Uprog) J p q r. 
       (WRITE Pr) CONF r /\
        REACH Pr J p q /\ |== (q IMP r) ==> REACH Pr J p r",
     REPEAT STRIP_TAC
     THEN REACH_TRANS_TAC "q:^Pred"
     THEN ASM_REWRITE_TAC[]
     THEN IMP_RES_TAC REACH_IMP_CONF
     THEN IMP_RES_TAC REACH_IMP_UNITY
     THEN IMP_RES_TAC REACH_IMP_STABLE
     THEN IMP_RES_TAC REACH_IMP_LIFT) ;;

let REACH_gMONO = prove_thm
  (`REACH_gMONO`,
  "!(Pr:^Uprog) J p q r. 
       (WRITE Pr) CONF r /\
        REACH Pr J p q /\ |== ((q AND J) IMP r) ==> REACH Pr J p r",
     REPEAT STRIP_TAC
     THEN REACH_TRANS_TAC "q:^Pred"
     THEN ASM_REWRITE_TAC[]
     THEN IMP_RES_TAC REACH_IMP_CONF
     THEN IMP_RES_TAC REACH_IMP_UNITY
     THEN IMP_RES_TAC REACH_IMP_STABLE
     THEN IMP_RES_TAC REACH_IMP_gLIFT) ;;

let REACH_MONO_TAC p =
    MATCH_MP_TAC REACH_MONO
    THEN EXISTS_TAC p
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_ANTIMONO = prove_thm
  (`REACH_ANTIMONO`,
  "!(Pr:^Uprog) J p q r. 
        (WRITE Pr) CONF p /\
        REACH Pr J q r /\ |== (p IMP q) ==> REACH Pr J p r",
     REPEAT STRIP_TAC
     THEN REACH_TRANS_TAC "q:^Pred"
     THEN ASM_REWRITE_TAC[]
     THEN IMP_RES_TAC REACH_IMP_UNITY
     THEN IMP_RES_TAC REACH_IMP_CONF
     THEN IMP_RES_TAC REACH_IMP_STABLE
     THEN IMP_RES_TAC REACH_IMP_LIFT) ;;

let REACH_gANTIMONO = prove_thm
  (`REACH_gANTIMONO`,
  "!(Pr:^Uprog) J p q r. 
        (WRITE Pr) CONF p /\
        REACH Pr J q r /\ |== ((p AND J) IMP q) ==> REACH Pr J p r",
     REPEAT STRIP_TAC
     THEN REACH_TRANS_TAC "q:^Pred"
     THEN ASM_REWRITE_TAC[]
     THEN IMP_RES_TAC REACH_IMP_UNITY
     THEN IMP_RES_TAC REACH_IMP_CONF
     THEN IMP_RES_TAC REACH_IMP_STABLE
     THEN IMP_RES_TAC REACH_IMP_gLIFT) ;;

let REACH_ANTIMONO_TAC p =
    MATCH_MP_TAC REACH_ANTIMONO
    THEN EXISTS_TAC p
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let lemma1 = prove
  ("|== ((J:^Pred) IMP (p EQUAL r)) ==> |== ((r AND J) IMP p)",
    pred_JADE_TAC 1) ;;

let lemma2 = prove
  ("|== ((J:^Pred) IMP (q EQUAL s)) ==> |== ((q AND J) IMP s)",
    pred_JADE_TAC 1) ;;

let REACH_SUBST = prove_thm
  (`REACH_SUBST`,
   "!(Pr:^Uprog) J p q r s.
        (WRITE Pr) CONF r /\ (WRITE Pr) CONF s /\
         |== (J IMP (p EQUAL r)) /\ |== (J IMP (q EQUAL s)) /\
         REACH Pr J p q
         ==>
         REACH Pr J r s",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC lemma1
    THEN IMP_RES_TAC lemma2
    THEN IMP_RES_TAC REACH_gMONO
    THEN IMP_RES_TAC REACH_gANTIMONO) ;;

%---------------------------------------------------------------------------------
  REACH is reflexive. 
---------------------------------------------------------------------------------%

let REACH_REFL = prove_thm
  (`REACH_REFL`,
   "!(Pr:^Uprog) J p. 
       UNITY Pr /\ STABLE Pr J /\ (WRITE Pr) CONF p ==> REACH Pr J p p",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC REACH_IMP_LIFT
    THEN ASM_REWRITE_TAC[]
    THEN pred_JADE_TAC 1) ;;

%---------------------------------------------------------------------------------
                           Disjunctivity of REACH
---------------------------------------------------------------------------------%

let REACH_DISJ = prove_thm 
  (`REACH_DISJ`,
   "!(W:*ind->bool) (Pr:^Uprog) J. DISJ W (REACH Pr J)",
    REWRITE_TAC [DISJ_DEF]
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RATOR_CONV o RAND_CONV o REWRITE_CONV)
         [qOR_RES_CHANGE]
    THEN REACH_LDISJ_TAC
    THENL
    [ %--- prove non-emptiness ---%
      BETA_TAC 
      THEN CONV_TAC SWAP_EXISTS_CONV
      THEN EXISTS_TAC "i:*ind" THEN EXISTS_TAC "(f:*ind->^Pred) i"
      THEN ASM_REWRITE_TAC[] ;
      %--- prove disjunctivity ---%
      RESTRICT_DEF_TAC
      THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[] THEN RES_TAC
      THEN IMP_RES_TAC REACH_IMP_CONF
      THEN REACH_MONO_TAC "(g:*ind->^Pred) i'"
      THENL
      [ CONF_TAC 1
        THENL 
        [ EXISTS_TAC "i:*ind" THEN ASM_REWRITE_TAC[] ;
          RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
          THEN RES_TAC THEN IMP_RES_TAC REACH_IMP_CONF ] ;
        pred_JADE_TAC 1
        THEN EXISTS_TAC "i':*ind"
        THEN ASM_REWRITE_TAC[] ]
    ]) ;;  

let REACH_GEN_LDISJ = prove_thm
  (`REACH_GEN_LDISJ`,
   "!W (Pr:^Uprog) J f q. 
       (?i. W i) /\ (!i:*ind::W. REACH Pr J (f i) q) 
        ==>
        REACH Pr J (??i::W.f i) q",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC GEN_LDISJ
   THEN ASM_REWRITE_TAC [REACH_DISJ]
   THEN EXISTS_TAC "i:*ind" THEN ASM_REWRITE_TAC[] ) ;;

let REACH_GEN_LDISJ_TAC f =
   let thm = (GEN_ALL o BETA_RULE o (ISPEC f) o (GEN "f:*ind->^Pred") o
              SPEC_ALL)  REACH_GEN_LDISJ  in
   MATCH_MP_TAC thm
   THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;


let REACH_SIMPLE_DISJ = prove_thm
  (`REACH_SIMPLE_DISJ`,
   "!(Pr:^Uprog) J p q r s.
         REACH Pr J p q /\ REACH Pr J r s 
         ==> REACH Pr J (p OR r) (q OR s)",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC SIMPLE_DISJ
   THEN ASM_REWRITE_TAC [REACH_DISJ]) ;;

let REACH_SIMPLE_DISJ_TAC =
    MATCH_MP_TAC REACH_SIMPLE_DISJ
    THEN ASM_REWRITE_TAC[]
    THEN REPEAT CONJ_TAC ;;

let REACH_CANCEL = prove_thm
  (`REACH_CANCEL`,
   "!(Pr:^Uprog) J p q r s.
       (WRITE Pr) CONF q /\
       REACH Pr J p (q OR r) /\ REACH Pr J r s ==> REACH Pr J p (q OR s)",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CANCEL
    THEN EXISTS_TAC "r:^Pred"
    THEN ASM_REWRITE_TAC [REACH_TRANS; REACH_DISJ]
    THEN IMP_RES_TAC REACH_IMP_UNITY
    THEN IMP_RES_TAC REACH_IMP_STABLE
    THEN IMP_RES_TAC REACH_REFL) ;;

let REACH_CANCEL_TAC r =
    MATCH_MP_TAC REACH_CANCEL
    THEN EXISTS_TAC r THEN ASM_REWRITE_TAC[]
    THEN REPEAT CONJ_TAC ;;

%---------------------------------------------------------------------------------
   PSP Laws for REACH

   General :     
                      REACH J p q
               (a /\ K) UNLESS b, STABLE K
            ----------------------------------
              REACH (J/\K) (p/\a) ((q/\a)\/b)

   as corollary, standard PSP:       

                       REACH J p q
                     (a /\ J) UNLESS b
               --------------------------------
                 REACH (J) (p/\a) ((q/\a)\/b)

   another corollary is that we can strengthen the stability assumption of a
   progress by reach.
---------------------------------------------------------------------------------%

let lemma1 = prove
   ("((a AND b) AND (c AND d)) = ((b AND d) AND (a AND (c:^Pred)))",
     pred_CYAN_TAC) ;;

let REACH_GEN_PSP = prove_thm
  (`REACH_GEN_PSP`,
   "!(Pr:^Uprog) J L p q a b.
       (WRITE Pr) CONF a /\ (WRITE Pr) CONF b /\
       REACH Pr J p q /\ 
       UNLESS Pr (a AND L)  b /\ STABLE Pr L
       ==>
       REACH Pr (J AND L) (p AND a) ((q AND a) OR b)",
    REPEAT STRIP_TAC
    THEN CONV_TAC (MK_2ABS_CONV "p:^Pred" "q:^Pred")
    THEN UNDISCH_TAC "REACH (Pr:^Uprog) J p q"
    THEN REACH_INDUCT1_TAC
    THENL
    [ %--- ENSURES case ---%
      REACH_ENS_LIFT_TAC
      THEN IMP_RES_TAC ENSURES_PSP
      THEN CONF_TAC 2 
      THENL [ ALL_TAC ; IMP_RES_TAC STABLE_SIMPLE_CONJ ]
      THEN ONCE_REWRITE_TAC [lemma1]
      THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
      THEN EXISTS_TAC 
           "(((a AND L) AND y) OR ((x AND J) AND b)) OR (b AND (y:^Pred))"
      THEN ASM_REWRITE_TAC[]
      THEN pred_JADE_TAC 3 ;
      %--- transitivity ---%
      SUBST1_TAC (SYM(ISPEC "b:^Pred" pOR_REFL))
      THEN ONCE_REWRITE_TAC [SYM (SPEC_ALL pOR_ASSOC)]
      THEN ONCE_REWRITE_TAC [SYM (SPEC_ALL pOR_SYM)]
      THEN REACH_CANCEL_TAC "y AND (a:^Pred)"
      THEN ONCE_REWRITE_TAC [SYM (SPEC_ALL pOR_SYM)]
      THEN ASM_REWRITE_TAC[] ;
      %--- disjunctivity ---%
      (CONV_TAC o RATOR_CONV o RAND_CONV o RATOR_CONV o RAND_CONV o
            RAND_CONV o ABS_CONV o ONCE_REWRITE_CONV) 
            [SYM (ISPEC "p:^Pred" I_THM)]
       THEN REWRITE_TAC [qAND_RIGHT_DISTR_OR; I_THM]
       THEN REACH_GEN_LDISJ_TAC "\i:^Pred. i AND a"
       THEN EXISTS_TAC "x:^Pred" THEN ASM_REWRITE_TAC[] 
      ] ) ;;

let REACH_PSP = prove_thm
  (`REACH_PSP`,
   "!(Pr:^Uprog) J p q a b.
       (WRITE Pr) CONF a /\ (WRITE Pr) CONF b /\
       REACH Pr J p q /\ 
       UNLESS Pr (a AND J) b 
       ==>
       REACH Pr J (p AND a) ((q AND a) OR b)",
    REPEAT STRIP_TAC 
    THEN (SUBST1_TAC o SYM o ISPEC "J:^Pred") pAND_REFL
    THEN MATCH_MP_TAC REACH_GEN_PSP
    THEN IMP_RES_TAC REACH_IMP_STABLE
    THEN ASM_REWRITE_TAC[]) ;;

let REACH_STAB_MONO = prove_thm
  (`REACH_STAB_MONO`,
   "!(Pr:^Uprog) J H p q. 
       STABLE Pr H /\ REACH Pr J p q ==> REACH Pr (H AND J) p q",
    REPEAT STRIP_TAC 
    THEN (SUBST1_TAC o SYM o ISPEC "p:^Pred") pAND_UNIT
    THEN (SUBST1_TAC o SYM o ISPEC "q:^Pred") pAND_UNIT
    THEN (SUBST1_TAC o SYM o ISPEC "TT AND (q:^Pred)") pOR_UNIT  
    THEN ONCE_REWRITE_TAC [pAND_SYM] 
    THEN ONCE_REWRITE_TAC [pOR_SYM] 
    THEN MATCH_MP_TAC REACH_GEN_PSP
    THEN ASM_REWRITE_TAC [pAND_UNIT; SYM(SPEC_ALL STABLE)]
    THEN CONJ_TAC THEN CONF_TAC 1) ;;

let REACH_PSP_TAC =
    MATCH_MP_TAC REACH_PSP
    THEN ASM_REWRITE_TAC [] THEN REPEAT CONJ_TAC ;;

let REACH_STABLE_PSP = prove_thm
  (`REACH_STABLE_PSP`,
   "!(Pr:^Uprog) J p q a.
        (WRITE Pr) CONF a /\
        REACH Pr J p q /\ STABLE Pr ( a AND J)
        ==>
        REACH Pr J (p AND a) (q AND a)",
    REWRITE_TAC [STABLE] THEN REPEAT STRIP_TAC
    THEN SUBST1_TAC (SYM (ISPEC "(q:^Pred) AND a" pOR_UNIT))
    THEN ONCE_REWRITE_TAC [pOR_SYM]
    THEN MATCH_MP_TAC REACH_PSP
    THEN ASM_REWRITE_TAC[]
    THEN CONF_TAC 1) ;;


%---------------------------------------------------------------------------------
  STABLE Shift Law : a corollary from left shift principle.

          REACH Pr (J AND K) p q
              STABLE Pr J
      -----------------------------
          REACH Pr J (p /\ K) q
---------------------------------------------------------------------------------%

let lemma1 = prove
   ("((a AND b) AND c) = (a AND (c AND (b:^Pred)))",
     pred_CYAN_TAC) ;;

let REACH_STABLE_SHIFT = prove_thm
  (`REACH_STABLE_SHIFT`,
   "!(Pr:^Uprog) J a p q.
          (WRITE Pr) CONF a /\
          REACH Pr (J AND a) p q /\ STABLE Pr J
          ==>
          REACH Pr J (p AND a) q",
    REPEAT STRIP_TAC
    THEN CONV_TAC (MK_2ABS_CONV "p:^Pred" "q:^Pred")
    THEN UNDISCH_TAC "REACH (Pr:^Uprog) (J AND a) p q"
    THEN REACH_INDUCT3_TAC
    THENL  
    [ %--- ensures --%
      REACH_ENS_LIFT_TAC
      THEN XRULE_ASSUM_TAC (REWRITE_RULE [B_ENS])
      THEN ASM_REWRITE_TAC [lemma1] 
      THEN CONF_TAC 2 ;
      %-- trans --%
      REACH_TRANS_TAC "(q:^Pred) AND a"
      THEN (SUBST1_TAC o SYM o ISPEC "(q:^Pred) AND a") pOR_UNIT
      THEN ONCE_REWRITE_TAC [pOR_SYM]
      THEN SUBST1_TAC (prove ("(p AND a) = ((p AND a) AND (a:^Pred))",
                       pred_CYAN_TAC))
      THEN REACH_PSP_TAC
      THENL 
      [ CONF_TAC 1 ; 
        REWRITE_TAC [SYM(SPEC_ALL STABLE)] 
        THEN ONCE_REWRITE_TAC [pAND_SYM]
        THEN IMP_RES_TAC REACH_IMP_STABLE ] ;
      %--- disj ---%
      (CONV_TAC o RATOR_CONV o RAND_CONV o RATOR_CONV o RAND_CONV o
       RAND_CONV o ABS_CONV o ONCE_REWRITE_CONV) [SYM (ISPEC "p:^Pred" I_THM)]
       THEN REWRITE_TAC [qAND_RIGHT_DISTR_OR; I_THM]
       THEN REACH_GEN_LDISJ_TAC "\i:^Pred. i AND a"
       THEN EXISTS_TAC "p:^Pred"
       THEN ASM_REWRITE_TAC[] ]) ;;

%---------------------------------------------------------------------------------
                           REACH Simple Completion Law
---------------------------------------------------------------------------------%

let Type_Of_X = ":^Pred->^Pred->^Pred->^Pred->bool";;

let REACH_DOUBLE_INDUCT = prove
  ("!(X:^Type_Of_X) (Pr:^Uprog) J. 
       (!p q r s. X p q r s  = X r s p q)                  /\
       (!p q r s. B_ENS Pr J p q ==> X p q r s)             /\
       (!W q r s. 
            ~(W=FF) /\ (!p :: W. REACH Pr J p q /\  X p q r s) 
            ==> X (?? p :: W. p) q r s)                           /\
       (!p q r u v w.  
            B_ENS Pr J p q /\ B_ENS Pr J u v /\
            REACH Pr J q r /\ REACH Pr J v w /\
            X p r v w /\ X u w q r /\ X q r v w 
             ==> X p r u w)  
        ==>
       (!p q r s. REACH Pr J p q /\ REACH Pr J r s ==> X p q r s)",
    REWRITE_TAC [REACH; TDC_EQU_Ri_TDC; Non_FF_lemma;
                 Ri_TDC_DOUBLE_INDUCT]) ;;

let REACH_DOUBLE_INDUCT_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC REACH_DOUBLE_INDUCT
   THEN REWRITE_TAC [B_ENS]
   THEN BETA_TAC 
   THEN REPEAT STRIP_TAC ;;

let lemma1 = prove
    %--- this lemma is used to prove ENSURES case of REACH_COMPL_lemma1 ---%
  ("!(Pr:^Uprog) J p q b.
       UNLESS Pr (q AND J) b /\ UNLESS Pr (p AND J) q /\ STABLE Pr J 
       ==>
       UNLESS Pr ((p OR q) AND J) b",
    REWRITE_TAC [STABLE]
    THEN REDUCE_UNLESS_TAC
    THENL
    [ ASM_CASES_TAC "(q:^Pred) s" ; 
      ASM_CASES_TAC "(p:^Pred) s" ]
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[])) ;;
      
let lemma2 = prove
    %--- this lemma is used to prove ENSURES case of REACH_COMPL_lemma1 ---%
  ("!(Pr:^Uprog) J p q r s b.
        (WRITE Pr) CONF b /\ 
        UNLESS Pr ((p OR q) AND J) b /\
        UNLESS Pr (s AND J) b /\
        REACH Pr J p q /\ REACH Pr J r s
        ==>
        REACH Pr J (p AND r) ((q AND s) OR b)",
    REPEAT STRIP_TAC
    THEN SUBST1_TAC (SYM(ISPEC "(q AND s) OR (b:^Pred)" pOR_REFL))
    THEN REACH_CANCEL_TAC "p AND (s:^Pred)"
    THENL
    [ IMP_RES_TAC REACH_IMP_CONF THEN CONF_TAC 3 ;
      REWRITE_TAC [prove
           ("(((q AND s) OR b) OR (p AND s)) = ((s AND (p OR q)) OR (b:^Pred))",
            pred_CYAN_TAC)]
      THEN REACH_ANTIMONO_TAC "(r:^Pred) AND ( p OR q)"
      THENL
      [ IMP_RES_TAC REACH_IMP_CONF THEN CONF_TAC 3 ;
        REACH_PSP_TAC THEN IMP_RES_TAC REACH_IMP_CONF THEN CONF_TAC 3 ;
        pred_JADE_TAC 3 ] ;
      REACH_PSP_TAC
      THEN IMP_RES_TAC REACH_IMP_CONF THEN CONF_TAC 3]) ;;


let lemma3 = prove
    %--- this lemma is used to prove trans. case in REACH_COMPL_lemma1 ---%
  ("!(Pr:^Uprog) J p q u v.
       ENSURES Pr (p AND J) q /\ ENSURES Pr (u AND J) v
       ==>
       ENSURES Pr ((p AND u) AND J) ((p AND v) OR (u AND q) OR (q AND v))",
    REPEAT STRIP_TAC
    THEN ONCE_REWRITE_TAC [pAND_ROVER_AND]
    THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
    THEN EXISTS_TAC 
         "(((p AND J) AND v)  OR ((u AND J) AND q)) OR (q AND (v:^Pred))"
    THEN CONJ_TAC
    THENL 
    [ OLD_IMP_RES_TAC ENSURES_CONJ ;
      pred_JADE_TAC 3 ]) ;;

let lemma4 = prove
    %--- this lemma is used to prove trans. case in REACH_COMPL_lemma1 ---%
  ("!(Pr:^Uprog) J p q r u v w b.
        REACH Pr J (p AND v) ((r AND w) OR b) /\
        REACH Pr J (u AND q) ((w AND r) OR b) /\
        REACH Pr J (q AND v) ((r AND w) OR b) /\
        REACH Pr J (p AND u) ((p AND v) OR (u AND q) OR (q AND v))
        ==>
        REACH Pr J (p AND u) ((r AND w) OR b)",
    REPEAT GEN_TAC
    THEN SUBST1_TAC (ISPECL ["w:^Pred"; "r:^Pred"] pAND_SYM)
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC REACH_IMP_CONF    
    THEN SUBST1_TAC (SYM (ISPEC "(r AND w) OR (b:^Pred)" pOR_REFL))
    THEN REACH_CANCEL_TAC "p AND (v:^Pred)"
    THEN (CONV_TAC o RAND_CONV o ONCE_REWRITE_CONV) [pOR_SYM]
    THEN SUBST1_TAC (SYM (ISPEC "(r AND w) OR (b:^Pred)" pOR_REFL))
    THEN ONCE_REWRITE_TAC [SYM(SPEC_ALL pOR_ASSOC)]
    THEN REACH_CANCEL_TAC "u AND (q:^Pred)"
    THENL [ CONF_TAC 5 ; ALL_TAC ]
    THEN ONCE_REWRITE_TAC [pOR_ASSOC]
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ONCE_REWRITE_CONV) [pOR_SYM]
    THEN ONCE_REWRITE_TAC [SYM(SPEC_ALL pOR_ASSOC)]
    THEN REACH_CANCEL_TAC "q AND (v:^Pred)"
    THENL [ CONF_TAC 5 ; ALL_TAC ]
    THEN ONCE_REWRITE_TAC [pOR_ASSOC]
    THEN ASM_REWRITE_TAC[]) ;;


let REACH_COMPL_lemma1 = prove
   %--- this lemma is used to prove REACH_COMPL_lemma2 ---%
  ("REACH (Pr:^Uprog) J p q /\ REACH Pr J r s ==>
       (\p q r s.
       !b.  (WRITE Pr) CONF b /\ 
            REACH Pr J p q /\ REACH Pr J r s /\ 
            UNLESS Pr (q AND J) b  /\ UNLESS Pr (s AND J) b
            ==> REACH Pr J (p AND r) ((q AND s) OR b)
        )p q r s",
    REACH_DOUBLE_INDUCT_TAC
    THENL
    [ %--- commutativity ---%
      EQ_TAC THEN REPEAT STRIP_TAC 
      THEN ONCE_REWRITE_TAC [pAND_SYM] 
      THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
      %--- ENSURES case ---%
      MATCH_MP_TAC lemma2
      THEN ASM_REWRITE_TAC[]
      THEN XRULE_ASSUM_TAC (REWRITE_RULE [ENSURES])
      THEN MATCH_MP_TAC lemma1
      THEN ASM_REWRITE_TAC[] ;
      %--- disjunctivity ---%
      REWRITE_TAC [(BETA_RULE o ISPECL ["W:^Pred->bool"; "\p:^Pred. p"])
                    qAND_RIGHT_DISTR_OR]
      THEN REACH_GEN_LDISJ_TAC "\i:^Pred. i AND a"
      THENL
      [ ASM_REWRITE_TAC [SYM (SPEC_ALL Non_FF_lemma)] ;
        UNDISCH_ALL_TAC 
        THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
        THEN RES_TAC ] ;
      %--- transitivity ---%
      MATCH_MP_TAC lemma4
      THEN EXISTS_TAC "q:^Pred" THEN EXISTS_TAC "v:^Pred"
      THEN REPEAT CONJ_TAC
      THENL 
      [ FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
        FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;        
        FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
        REACH_ENS_LIFT_TAC
        THEN CONF_TAC 4
        THEN MATCH_MP_TAC lemma3 THEN ASM_REWRITE_TAC[] ]
   ] ) ;;


let REACH_COMPL_lemma2 = prove
  %--- this lemma is used to prove REACH_SIMPLE_COMPLETION ---%
  ("!(Pr:^Uprog) p q r s b.
       (WRITE Pr) CONF  b /\
        REACH Pr J p q /\ REACH Pr J r s /\
        UNLESS Pr (q AND J) b /\ UNLESS Pr (s AND J) b
        ==> 
        REACH Pr J (p AND r) ((q AND s) OR b)",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (BETA_RULE REACH_COMPL_lemma1)) ;;


let lemma5 = prove
    %--- this lemma is used to prove REACH_SIMPLE_COMPLETION ---%
  ("!(Pr:^Uprog) J p b.
       UNLESS Pr (p AND J) b ==> UNLESS Pr ((p OR b) AND J) b",
    REDUCE_UNLESS_TAC
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[]));;


let REACH_SIMPLE_COMPLETION = prove_thm
  (`REACH_SIMPLE_COMPLETION`,
   "!(Pr:^Uprog) p q r s b. 
        (WRITE Pr) CONF b /\ 
         REACH Pr J p (q OR b) /\ REACH Pr J r (s OR b) /\
         UNLESS Pr (q AND J) b /\ UNLESS Pr (s AND J) b ==>
         REACH Pr J (p AND r) ((q AND s) OR b)", 
   ONCE_REWRITE_TAC [ prove
        ("((q AND s) OR b) = (((q OR b) AND (s OR b)) OR (b:^Pred))",
          pred_CYAN_TAC)]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC REACH_COMPL_lemma2
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC THEN IMP_RES_TAC lemma5) ;;

%---------------------------------------------------------------------------------
                            REACH Completion Law
---------------------------------------------------------------------------------%

let MK_ABS_CONV x f = SYM (BETA_CONV "(\^x. ^f) ^x") ;;

let lemma6 = 
    (GEN_ALL o BETA_RULE)
    (ISPECL ["Pr:^Uprog"; "W:**->bool";
            "\i:**. (P i) AND (J:^Pred)"; "\i:**. b:^Pred"]
     UNLESS_GEN_CONJ) ;;

let UNLESS_CONJ_lemma = prove
  ("!(Pr:^Uprog) J Q b.
      (!i:**::V.  UNLESS (Pr:^Uprog) ((Q i) AND J) b) /\
       STABLE Pr J 
      ==> 
      UNLESS Pr ((!!i::V.Q i) AND J) b",
    REPEAT STRIP_TAC
    THEN ASM_CASES_TAC "(V:**->bool)=FF"
    THENL 
    [ ASM_REWRITE_TAC [qAND_EMPTY; pAND_UNIT] 
      THEN UNDISCH_TAC "STABLE (Pr:^Uprog) J"
      THEN REWRITE_TAC [STABLE]
      THEN REDUCE_UNLESS_TAC
      THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[]) ;
      FIRST_ASSUM (\thm. REWRITE_TAC [MATCH_MP qAND_RIGHT_DISTR_AND thm])
      THEN MATCH_MP_TAC UNLESS_CONSQ_WEAK
      THEN EXISTS_TAC "(!!i:**::V. ((Q i) AND J) OR b) AND (??i::V. (b:^Pred))"
      THEN CONJ_TAC
      THENL 
      [ MATCH_MP_TAC lemma6 THEN ASM_REWRITE_TAC[] ;
        pred_JADE_TAC 3]]) ;;       


let REACH_COMPL_lemma3 = prove
    %--- this lemma is used to prove REACH_COMPLETION ---%
  ("!(W:**->bool)::CHF_FINITE.
    ! (Pr:^Uprog) J P Q b. 
        UNITY Pr /\ 
        STABLE Pr J /\
       (WRITE Pr) CONF b /\ 
       (!i :: W. REACH Pr J (P i) ((Q i) OR b)) /\
       (!i :: W. UNLESS Pr ((Q i) AND J) b) 
        ==>
        REACH Pr J (!! i :: W. P i) ((!! i :: W. Q i) OR b)",
    RESTRICT_DEF_TAC
    THEN REPEAT_FIN 7 STRIP_TAC
    THEN CONV_TAC (MK_ABS_CONV "x:**->bool")
    THEN UNDISCH_TAC "CHF_FINITE (x:**->bool)"
    THEN MATCH_MP_TAC CHF_INDUCTION
    THEN BETA_TAC THEN REPEAT STRIP_TAC 
    THENL
    [ %--- base case: W = FF ---%
      REWRITE_TAC [qAND_EMPTY; pOR_ZERO]
      THEN MATCH_MP_TAC REACH_REFL
      THEN ASM_REWRITE_TAC[] THEN CONF_TAC 1 ;
      %--- induction case ---%
      XRULE_ASSUM_TAC BETA_RULE
      THEN ONCE_REWRITE_TAC [DISJ_SYM]
      THEN ONCE_REWRITE_TAC [EQ_SYM_EQ]
      THEN REWRITE_TAC [SYM (SPEC_ALL qAND_DOM_SPLIT)]
      THEN MATCH_MP_TAC REACH_SIMPLE_COMPLETION
      THEN REPEAT CONJ_TAC
      THENL 
      [ ASM_REWRITE_TAC[] ;
        FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC 
        THEN ASM_REWRITE_TAC [] ;
        FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[] ;
        MATCH_MP_TAC UNLESS_CONJ_lemma   
        THEN RESTRICT_DEF_TAC THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC 
        THEN FIRST_ASSUM MATCH_MP_TAC 
        THEN ASM_REWRITE_TAC[] ;
        FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[] ]
   ]) ;;


let REACH_COMPLETION = prove_thm
  (`REACH_COMPLETION`,
   "!(W:**->bool) (Pr:^Uprog) J P Q b. 
        CHF_FINITE W /\ UNITY Pr /\
       (WRITE Pr) CONF b /\ STABLE Pr J /\
       (!i :: W. REACH Pr J (P i) ((Q i) OR b)) /\
       (!i :: W. UNLESS Pr ((Q i) AND J) b) 
        ==>
        REACH Pr J (!! i :: W. P i) ((!! i :: W. Q i) OR b)",
    RESTRICT_DEF_TAC 
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (QUAFF_RULE 3  REACH_COMPL_lemma3)) ;;

%---------------------------------------------------------------------------------
                          REACH-STABLE conjunction
---------------------------------------------------------------------------------%

let REACH_STABLE_SIMPLE_CONJ = prove_thm
  (`REACH_STABLE_SIMPLE_CONJ`,
   "!(Pr:^Uprog) p q r s.
        REACH Pr J p q /\ REACH Pr J r s /\
        STABLE Pr (q AND J) /\ STABLE Pr (s AND J) 
        ==> 
        REACH Pr J (p AND r) (q AND s)",
    REWRITE_TAC [STABLE]
    THEN REPEAT STRIP_TAC
    THEN SUBST1_TAC (SYM (ISPEC "q AND (s:^Pred)" pOR_UNIT))
    THEN ONCE_REWRITE_TAC [pOR_SYM]
    THEN MATCH_MP_TAC REACH_COMPL_lemma2
    THEN ASM_REWRITE_TAC[]
    THEN CONF_TAC 1) ;;

let REACH_STABLE_CONJ = prove_thm
  (`REACH_STABLE_CONJ`,
   "!(W:**->bool) (Pr:^Uprog) J P Q. 
        CHF_FINITE W /\ UNITY Pr /\ STABLE Pr J /\
       (!i::W. REACH Pr J (P i) (Q i)) /\
       (!i::W. STABLE Pr ((Q i) AND J))
        ==>
        REACH Pr J (!!i::W. P i) (!!i::W. Q i)",
    REWRITE_TAC [STABLE]
    THEN REPEAT STRIP_TAC
    THEN SUBST1_TAC (SYM (ISPEC "!!i:**::W. ((Q i):^Pred)" pOR_UNIT))
    THEN ONCE_REWRITE_TAC [pOR_SYM]
    THEN MATCH_MP_TAC REACH_COMPLETION
    THEN ASM_REWRITE_TAC[STABLE]
    THEN REPEAT CONJ_TAC
    THENL
    [ CONF_TAC 1 ;
      ONCE_REWRITE_TAC [pOR_SYM]
      THEN ASM_REWRITE_TAC [pOR_UNIT]]) ;;


close_theory();;

%---------------------------------------------------------------------------------
   written by:

     Wishnu Prasetya
     Dept. of Computer Science
     Utrecht University
     the Netherlands

     email: wishnu@cs.ruu.nl
---------------------------------------------------------------------------------%   
