%----------------------------------------------------------------------------------
 This theory defines what a UNITY program is and defines the basic UNITY 
 operators.

 A UNITY program is represented by a quadruple (A,J,R,W) where:

      A: a predicate representing the set of actios
      J: a predicate describing possible initial states
      R: a predicate describing the set of read variables
      W: a predicate describing the set of write variables

 We assume a global, but further left unspecified, set of available program 
 variables and their values. Here it is represented it by polymorphic types
 *var and *value. 

 An action is represented as a relation on State (= *var->*value). That is,
 an action a has the type State->State->bool. If a is an action, a s describes 
 the possible states after executing a in state s. So there is non-determinism.

 Definitions provided by this theory:

    UNITY     : defining what a UNITY program is
    HOA       : Hoare Triple
    UNLESS    : basic UNITY operator to express safety
    STABLE    : operator to express stability
    Inv       : operator to express (strong) invariant
    ENSURES   : basic UNITY operator to express progress
    LEADSTO   : a more general operator for progress
    LEADSTOs  : Sander's lead-to definition 
    UNL, ENS, LTO : Sander's subscripted version of UNLESS, ENSURES, and LEADSTO
    PAR       : UNITY parallel composition (program union)

 This theory also includes some basic properties of PAR such as its commutativity,
 associativity, etc.
      
 Author:

     Wishnu Prasetya
     Dept. of Computer Science
     Utrecht University
     the Netherlands

     email: wishnu@cs.ruu.nl

  Libraries required: WF (and relation, predicates, and sets)
----------------------------------------------------------------------------------%

load_library `WF` ;;
load_theory `TDC` ;;
new_theory `def_UNITY`;;
new_parent `variables` ;;
loadf `my_misc` ;;
loadf `OLD_RES` ;;

let Autoload thy =     
    let defs = map fst (definitions thy) in
    map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
    map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;
 
Autoload `relation` ;;
Autoload `TDC` ;;
Autoload `variables` ;;

let State = ":*var->*value" ;;
let Pred = ":(*var->*value)->bool" ;;
let Action = ":(*var->*value)->(*var->*value)->bool" ;;
let ACTIONS = ":((*var->*value)->(*var->*value)->bool)->bool" ;;
let VARS = ":*var->bool" ;;
let Uprog = ":^ACTIONS # ^Pred # ^VARS # ^VARS" ;;

%----------------------------------------------------------------------------------
  an action is called always enabled if its execution in any state s always leads
  to some state (so there is no issue of abortion or non-termination).
----------------------------------------------------------------------------------%

let ALWAYS_ENABLED = new_definition
  (`ALWAYS_ENABLED`, 
   "ALWAYS_ENABLED (A:^Action) = !s. ?t. A s t") ;;

%----------------------------------------------------------------------------------
   A quadruple (A,J,R,W) is a UNITY program iff:
   
     (1) A is non-empty
     (2) every action in A is always enabled
     (3) every action in A only reads from R
     (4) every action in A only write to R
     (5) W is included in R

----------------------------------------------------------------------------------%

let UNITY = new_definition 
  (`UNITY`,
   "UNITY (P:^ACTIONS, In:^Pred, R:^VARS, W:^VARS) = 
      (?A. P A) 
      /\ (!A::P. ALWAYS_ENABLED A)
      /\ (!A::P. (NOT W) IG_BY A)
      /\ (!x. W x ==> R x)
      /\ (!A::P. (NOT R) INVI A)") ;;

%----------------------------------------------------------------------------------
   Introduce destructors PROG, INIT, READ, and WRITE to access each component
   of a program.
----------------------------------------------------------------------------------%

let PROG = new_definition (`PROG`, "PROG (P:*,In:**,R:***,W:****) = P") ;;
let INIT = new_definition (`INIT`, "INIT (P:*,In:**,R:***,W:****) = In") ;;
let READ = new_definition (`READ`, "READ (P:*,In:**,R:***,W:****) = R") ;;
let WRITE = new_definition (`WRITE`, "WRITE (P:*,In:**,R:***,W:****) = W") ;;

let ID_UPROG = prove_thm
  (`ID_UPROG`, 
   "!Pr:^Uprog. Pr = (PROG Pr, INIT Pr, READ Pr, WRITE Pr)",
    GEN_TAC 
    THEN (SUBST1_TAC  o SYM o ISPEC "Pr:^Uprog") PAIR
    THEN (SUBST1_TAC  o SYM o ISPEC "SND (Pr:^Uprog)") PAIR
    THEN (SUBST1_TAC  o SYM o ISPEC "SND (SND (Pr:^Uprog))") PAIR
    THEN REWRITE_TAC [PROG; INIT; READ; WRITE]) ;;

%---------------------------------------------------------------------------------
   the input variables of a program is the non-writeable part of its read
   variables.
---------------------------------------------------------------------------------%

let INP = new_definition 
  (`INP`,
   "INP (Pr:^Uprog) = (NOT (WRITE Pr)) AND (READ Pr)") ;;

%----------------------------------------------------------------------------------
  The following definitions we will use later:
    
      - UP, for example UP V $= X is a predicate describing states s where 
            the value of each variable x in V is X x.

      - DVa Pr Qr, dependent variable. Describes the set of variables read by Pr,
            and written by Qr.   
----------------------------------------------------------------------------------%

let UP_DEF = new_definition (`UP_DEF`,
   "UP V R (C:*->**) = (\s:*->**. (!x. V x ==> R (s x) (C x)))") ;;

let DVa = new_definition (`DVa`,
   "DVa (Pr:^Uprog) (Qr:^Uprog) = (READ Pr) AND (WRITE Qr)") ;;

%----------------------------------------------------------------------------------
                                just some lemmas
----------------------------------------------------------------------------------%

let Non_FF_lemma = prove_thm
  (`Non_FF_lemma`,
   "!W:*->bool. ~(W=FF) = ?s. W s",
    REWRITE_TAC [X_FUN_EQ_CONV "s:*" "(W:*->bool)=FF"]
    THEN REWRITE_TAC [FF_DEF]
    THEN REWRITE_TAC [NOT_FORALL_CONV "~(!s:*. ~W s)"]) ;;

let FORALL_RES_CHANGE = prove_thm
  (`FORALL_RES_CHANGE`,
   "!(W:*->bool) f (P:*->**).
    (!i::W. f(P i)) = (!p::(\p'. ?i. (W i) /\ (p'=(P i))). f p)",
    RESTRICT_DEF_TAC THEN REPEAT GEN_TAC
    THEN EQ_TAC THEN (REPEAT STRIP_TAC)
    THENL [ RES_TAC THEN ASM_REWRITE_TAC [] ;
            RES_TAC THEN XRULE_ASSUM_TAC (SPEC "(P:*->**) x")
            THEN XRULE_ASSUM_TAC (REWRITE_RULE[])
            THEN ASM_REWRITE_TAC[]]) ;;

let qOR_RES_CHANGE = prove_thm
  (`qOR_RES_CHANGE`,
   "!(W:*ind->bool) (P:*ind->(^Pred)).
       (??i::W. P i) = (??p::(\p'. (?i. (W i) /\ (p'=(P i)))). p)",
    REWRITE_TAC [RES_qOR; pOR_DEF] THEN BETA_TAC
    THEN RESTRICT_DEF_TAC THEN REPEAT GEN_TAC
    THEN EXT_TAC THEN EQ_TAC 
    THEN REPEAT STRIP_TAC
    THENL [ EXISTS_TAC "(P:*ind->(^Pred)) i" THEN ASM_REWRITE_TAC []
            THEN EXISTS_TAC "i:*ind" THEN ASM_REWRITE_TAC [] ;
            EXISTS_TAC "i':*ind" 
            THEN XRULE_ASSUM_TAC SYM
            THEN ASM_REWRITE_TAC[] ] ) ;;

let Non_FF_DOM_CHANGE = prove_thm 
  (`Non_FF_DOM_CHANGE`,
   "!(W:*->bool) (P:*->(**->bool)).
        ~(W=FF) ==> ~((\x. ?i. (W i) /\ (x=(P i))) = FF)",
    REWRITE_TAC [Non_FF_lemma]
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN EXISTS_TAC "(P:*->(**->bool)) s" 
    THEN EXISTS_TAC "s:*" THEN ASM_REWRITE_TAC[]) ;;

%----------------------------------------------------------------------------------
                        definition of Hoare Triple
----------------------------------------------------------------------------------%

let HOA_DEF = new_definition 
   (`HOA_DEF`,
    "HOA(p,A:^Action,q) = !s t. (p s) /\ (A s t) ==> (q t)") ;;

%----------------------------------------------------------------------------------
               definition of UNITY primitive operators
----------------------------------------------------------------------------------%

let UNLESS = new_definition 
   (`UNLESS`,
    "UNLESS (Pr:^Uprog) p q = !A::PROG Pr. HOA(p AND (NOT q), A, p OR q)");;

let STABLE = new_definition
  (`STABLE`,
   "STABLE (Pr:^Uprog) p = UNLESS Pr p FF") ;;

let Inv = new_definition (`Inv`,
   "Inv (Pr:^Uprog) J = 
    ((|== ((INIT Pr) IMP J))  /\ UNLESS Pr J FF)") ;;

let ENSURES = new_definition
  (`ENSURES`,
   "ENSURES (Pr:^Uprog) p q = 
   (UNITY Pr) /\ (UNLESS Pr p q) /\ (?A::PROG Pr. HOA(p AND (NOT q),A,q))");;  

let LEADSTO = new_definition
  (`LEADSTO`,
   "LEADSTO (Pr:^Uprog) = TDC (ENSURES Pr)") ;;    

let PAR = new_infix_definition
  (`PAR`,
   "PAR (Pr:^Uprog) (Qr:^Uprog) = 
   ((PROG Pr) OR (PROG Qr), (INIT Pr) AND (INIT Qr), 
    (READ Pr) OR (READ Qr), (WRITE Pr) OR (WRITE Qr))") ;;

%----------------------------------------------------------------------------------
         definition of Sander's UNITY operators
----------------------------------------------------------------------------------%

let UNL = new_definition 
  (`UNL`,
   "UNL (Pr:^Uprog) J p q = UNLESS Pr (p AND J) (q AND J) /\ Inv Pr J") ;;

let ENS = new_definition 
  (`ENS`,
   "ENS (Pr:^Uprog) J p q = ENSURES Pr (p AND J) (q AND J) /\ Inv Pr J") ;;

let LTO = new_definition 
  (`LTO`,
   "LTO (Pr:^Uprog) J = TDC (ENS Pr J)") ;;

let ENSURESs = new_definition 
  (`ENSURESs`,
   "ENSURESs (Pr:^Uprog) p q = (?J. ENS Pr J p q)") ;;

let LEADSTOs = new_definition
  (`LEADSTOs`,
   "LEADSTOs (Pr:^Uprog) = TDC (ENSURESs Pr)") ;;


%----------------------------------------------------------------------------------
   Some basic properties of parallel composition. First of all, PAR is closed   
   in the space of UNITY programs.
----------------------------------------------------------------------------------%

let UNITY_PAR = prove_thm(`UNITY_PAR`,
   "!Pr (Qr:^Uprog). (UNITY Pr) /\ (UNITY Qr) ==>
                     (UNITY (Pr PAR Qr))",
    REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (ONCE_REWRITE_RULE [ID_UPROG])
    THEN UNDISCH_ALL_TAC
    THEN REWRITE_TAC [PAR; pOR_DEF; UNITY; PROG; READ; WRITE]
    THEN RESTRICT_DEF_TAC THEN BETA_TAC
    THEN REPEAT STRIP_TAC 
    THENL 
    [ EXISTS_TAC "A:^Action" THEN ASM_REWRITE_TAC[] ;
      RES_TAC ; RES_TAC ;
      MATCH_MP_TAC IG_BY_MONO
      THEN EXISTS_TAC "NOT(WRITE (Pr:^Uprog))"
      THEN CONJ_TAC
      THENL [ pred_JADE_TAC 3 ; RES_TAC ] ;
      MATCH_MP_TAC IG_BY_MONO
      THEN EXISTS_TAC "NOT(WRITE (Qr:^Uprog))"
      THEN CONJ_TAC
      THENL [ pred_JADE_TAC 3 ; RES_TAC ] ;
      RES_TAC THEN ASM_REWRITE_TAC[] ;
      RES_TAC THEN ASM_REWRITE_TAC[] ;
      MATCH_MP_TAC INVI_MONO
      THEN EXISTS_TAC "NOT(READ (Pr:^Uprog))"
      THEN REPEAT CONJ_TAC
      THENL 
      [ pred_JADE_TAC 3 ; 
        MATCH_MP_TAC IG_BY_MONO
        THEN EXISTS_TAC "NOT(WRITE (Pr:^Uprog))"
        THEN CONJ_TAC
        THENL [ pred_JADE_TAC 3 ; RES_TAC ] ;
        RES_TAC ] ;
      MATCH_MP_TAC INVI_MONO
      THEN EXISTS_TAC "NOT(READ (Qr:^Uprog))"
      THEN REPEAT CONJ_TAC
      THENL 
      [ pred_JADE_TAC 3 ; 
        MATCH_MP_TAC IG_BY_MONO
        THEN EXISTS_TAC "NOT(WRITE (Qr:^Uprog))"
        THEN CONJ_TAC
        THENL [ pred_JADE_TAC 3 ; RES_TAC ] ;
        RES_TAC ] ]) ;;

%----------------------------------------------------------------------------------
   PAR has a unit element, although this unit is not a UNITY program.
----------------------------------------------------------------------------------%

let PAR_UNIT = prove_thm
  (`PAR_UNIT`,
   "!(Pr:^Uprog). Pr PAR (FF,TT,FF,FF) = Pr",
    REWRITE_TAC [PAR; PROG; INIT; READ; WRITE]
    THEN ONCE_REWRITE_TAC [pOR_SYM; pAND_SYM]
    THEN REWRITE_TAC [pOR_UNIT; pAND_UNIT; SYM(SPEC_ALL ID_UPROG)]) ;;

%----------------------------------------------------------------------------------
             PAR is commutative and associative
----------------------------------------------------------------------------------%

let PAR_SYM = prove_thm (`PAR_SYM`,
   "!Pr (Qr:^Uprog). (Pr PAR Qr) = (Qr PAR Pr)",
    REWRITE_TAC [PAR] THEN REPEAT GEN_TAC
    THEN SUBST1_TAC (ISPECL ["PROG (Pr:^Uprog)";"PROG (Qr:^Uprog)"]
                     pOR_SYM)
    THEN SUBST1_TAC (ISPECL ["INIT (Pr:^Uprog)";"INIT (Qr:^Uprog)"]
                     pAND_SYM)
    THEN SUBST1_TAC (ISPECL ["READ (Pr:^Uprog)";"READ (Qr:^Uprog)"]
                     pOR_SYM)
    THEN SUBST1_TAC (ISPECL ["WRITE (Pr:^Uprog)";"WRITE (Qr:^Uprog)"]
                     pOR_SYM)
    THEN REWRITE_TAC[]) ;;

let PAR_ASSOC = prove_thm
  (`PAR_ASSOC`,
   "!(Pr:^Uprog) Qr Rr. (Pr PAR Qr) PAR Rr = Pr PAR (Qr PAR Rr)",
   REWRITE_TAC [PAR; PROG; INIT; READ; WRITE]
   THEN REWRITE_TAC [pAND_ASSOC; pOR_ASSOC]) ;;

%----------------------------------------------------------------------------------
   A predicate confined by a component program is also confined by the composite
   program.
----------------------------------------------------------------------------------%

let CONF_PAR = prove_thm
  (`CONF_PAR`,
   "!(Pr:^Uprog) Qr (p:^Pred).
          (WRITE Pr) CONF p \/ (WRITE Qr) CONF p
          ==> (WRITE (Pr PAR Qr)) CONF p",
    REWRITE_TAC [CONF_DEF; Pj_EQ; PAR; WRITE; pOR_DEF]
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC 
    THEN REPEAT STRIP_TAC THEN RES_TAC) ;;


close_theory();;
