%------------------------------------------------------------------------

  FILE		: mk_CLa.ml                                                

  A theory of complete lattice. 

  The theory is based on the definitions in CPO.th.

  LOADS LIBRARY	: sets, tauts			       		        
  READS FILES	: CPO.th
  WRITES FILES  : CLa.th					
 									
  AUTHOR	: Wishnu Prasetya
  AFFILIATION   : Dept. of Comp. Science, Utrecht University, 
                  the Netherlands
  DATE		: May 1995				

  Let in the sequel r be a relation on type A, s a relation on type B, 
  f a function from A to B, and X a set on type A.
  Concepts/definitions supported:


Lattices:
  CAP_CLa r    : r is a complete CAP  lattice
  CAP_PLa r    : r is a pointed CAP  lattice
  analog with CUP_CLa and CUP_PLa
     
  CLa r        : r is a complete lattice
 
Distributivity of a function over CUP and CAP:

  CUP_Junc r s f   : f distributes over CUP (least upper bound) of
                     any non-empty set X
  CAP_Junc r s f   : f distributes over CAP (greatest lower bound) of
                     any non-empty set X
  CUP_Distr r s f  : f distributes over CUP of any set X
  CAP_Distr r s f  : f distributes over CAP of any set X

Main results supported by this theory:

  - Knaster-Tarski fix point theorem for monotonic functions.
  - Rolling Rule of fix point.
  - Plotkin's lemma for fix point fusion of monotonic functions.

version2: Oct 94.
Wishnu Prasetya
wishnu@cs.ruu.nl

------------------------------------------------------------------------%

library_loader (`CLa`, [`sets`; `relation`; `taut`],
                       [`CPO`],[`my_misc`],``,``,[]) ;;
new_theory `CLa`;;

Autoload `relation` ;;

%-------------------------------------------------------------------------
  A complete CAP  lattice (CAP_CLa) is a partial ordering r 
  (over type A) such that the CAP of any non-empty set (over A) exists.

  A complete CAP pointed lattice (CAP_PLa) is the same as CAP  lattice
  except that the CAP of the empty set also exists.

  In the same way we can define complete CUP  lattices and CUP pointed
  lattices.

  A complete lattice (CLa) is an ordering that is both CAP and CUP pointed
  lattice.
-------------------------------------------------------------------------%

let CAP_CLa = new_definition
  (`CAP_CLa`,
   "CAP_CLa r = PO r /\ (!X. ~(X={}) ==> (?b:*. isCAP r X b))") ;;  

let CUP_CLa = new_definition
  (`CUP_CLa`,
   "CUP_CLa r = PO r /\ (!X. ~(X={}) ==> (?b:*. isCUP r X b))") ;;  

let CAP_PLa = new_definition
  (`CAP_PLa`,
   "CAP_PLa r = PO r /\ (!X. ?b:*. isCAP r X b)") ;;  

let CUP_PLa = new_definition
  (`CUP_PLa`,
   "CUP_PLa r = PO r /\ (!X. ?b:*. isCUP r X b)") ;; 

let CLa = new_definition
  (`CLa`,
   "CLa (r:*->*->bool) = CAP_PLa r /\ CUP_PLa r") ;;

%-------------------------------------------------------------------------
   If r is a complete lattice, then r is also a CAP and CUP  lattice
   (and vice versa).
-------------------------------------------------------------------------%

let CLa_ADEF = prove_thm
  (`CLa_ADEF`,
   "!r. CLa (r:*->*->bool) = CAP_CLa r /\ CUP_CLa r",
    GEN_TAC THEN REWRITE_TAC [CLa]
    THEN REWRITE_TAC [CAP_PLa; CUP_PLa; CAP_CLa; CUP_CLa]
    THEN EQ_TAC THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THEN ASM_CASES_TAC "(X:(*)set)={}"
    THENL
    [ ASM_REWRITE_TAC [isCAP_isCUP_EXTREMES2]
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN REWRITE_TAC [UNIV_NOT_EMPTY] ;
      FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
      ASM_REWRITE_TAC [SYM (SPEC_ALL isCAP_isCUP_EXTREMES1)]
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN REWRITE_TAC [UNIV_NOT_EMPTY] ;
      FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] 
    ]) ;;

%-------------------------------------------------------------------------
  A pointed CAP (CUP) lattice is also a CAP (CUP)  lattice.
-------------------------------------------------------------------------%

let CAP_PLa_IMP_CAP_CLa = prove_thm
  (`CAP_PLa_IMP_CAP_CLa`,
   "!r:*->*->bool. CAP_PLa r ==> CAP_CLa r",
    REWRITE_TAC [CAP_PLa; CAP_CLa] THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]) ;;

let CUP_PLa_IMP_CUP_CLa = prove_thm
  (`CUP_PLa_IMP_CUP_CLa`,
   "!r:*->*->bool. CUP_PLa r ==> CUP_CLa r",
    REWRITE_TAC [CUP_PLa; CUP_CLa] THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
  Conversely, a complete CAP (CUP) lattice with Top (Bottom) is also a 
  CAP (CUP)  pointed lattice.
-------------------------------------------------------------------------%

let CAP_CLa_IMP_PLa = prove_thm
  (`CAP_CLa_IMP_PLa`,
   "(?a:*. isCAP r {} a) /\ CAP_CLa r ==> CAP_PLa r",
    REWRITE_TAC [CAP_CLa; CAP_PLa]
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THEN ASM_CASES_TAC "(X:(*)set) = {}"
    THENL 
    [ EXISTS_TAC "a:*" THEN ASM_REWRITE_TAC[] ;
      FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ]) ;;

let CUP_CLa_IMP_PLa = prove_thm
  (`CUP_CLa_IMP_PLa`,
   "(?a:*. isCUP r {} a) /\ CUP_CLa r ==> CUP_PLa r",
    REWRITE_TAC [CUP_CLa; CUP_PLa]
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THEN ASM_CASES_TAC "(X:(*)set) = {}"
    THENL 
    [ EXISTS_TAC "a:*" THEN ASM_REWRITE_TAC[] ;
      FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ]) ;;

%-------------------------------------------------------------------------
  In a CUP (CAP)  lattice, the CUP (CAP) of a non-empty X is indeed 
  the least upper bound (greatest lower bound).

  Analog with pointed lattice.
-------------------------------------------------------------------------%

let CUP_CLa_CUP = prove_thm
  (`CUP_CLa_CUP`,
   "!r (X:(*)set). CUP_CLa r /\ ~(X={}) ==> isCUP r X (CUP r X)",
    REWRITE_TAC [CUP_CLa] THEN REPEAT STRIP_TAC
    THEN RES_TAC
    THEN IMP_RES_TAC CUP) ;;

let CAP_CLa_CAP = prove_thm
  (`CAP_CLa_CAP`,
   "!r (X:(*)set). CAP_CLa r /\ ~(X={}) ==> isCAP r X (CAP r X)",
    REWRITE_TAC [CAP_CLa] THEN REPEAT STRIP_TAC
    THEN RES_TAC
    THEN IMP_RES_TAC CAP) ;;

let CUP_PLa_CUP = prove_thm
  (`CUP_PLa_CUP`,
   "!r (X:(*)set). CUP_PLa r  ==> isCUP r X (CUP r X)",
    REWRITE_TAC [CUP_PLa] THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CUP
    THEN ASM_REWRITE_TAC[]) ;;

let CAP_PLa_CAP = prove_thm
  (`CAP_PLa_CAP`,
   "!r (X:(*)set). CAP_PLa r  ==> isCAP r X (CAP r X)",
    REWRITE_TAC [CAP_PLa] THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CAP
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
  In a pointed CAP (CUP) lattice, Bot (Top) is indeed the bottom (top)
  element. Both conclusions hold in a complete lattice.
-------------------------------------------------------------------------%

let CAP_PLa_Bot_THM = prove_thm
  (`CAP_PLa_Bot_THM`,
   "!r (x:*). CAP_PLa r ==> r (Bot r) x",
    REWRITE_TAC [CAP_PLa] THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Bot_THM
    THEN ASM_REWRITE_TAC[]) ;;

let CUP_PLa_Top_THM = prove_thm
  (`CUP_PLa_Top_THM`,
   "!r (x:*). CUP_PLa r ==> r x (Top r)",
    REWRITE_TAC [CUP_PLa] THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Top_THM
    THEN ASM_REWRITE_TAC[]) ;;

let CLa_Bot_THM = prove_thm
  (`CLa_Bot_THM`,
   "!r (x:*). CLa r ==> r (Bot r) x",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CAP_PLa_Bot_THM
    THEN ASM_REWRITE_TAC[]) ;;

let CLa_Top_THM = prove_thm
  (`CLa_Top_THM`,
   "!r (x:*). CLa r ==> r x (Top r)",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CUP_PLa_Top_THM
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
   In a complete lattice, the CUP (CAP) is indeed the least upper bound
   (greatest lower bound). 
-------------------------------------------------------------------------%

let CLa_CUP = prove_thm
  (`CLa_CUP`,
   "!r (X:(*)set). CLa r ==> isCUP r X (CUP r X)",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CUP_PLa_CUP THEN ASM_REWRITE_TAC[]) ;;

let CLa_CAP = prove_thm
  (`CLa_CAP`,
   "!r (X:(*)set). CLa r ==> isCAP r X (CAP r X)",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CAP_PLa_CAP THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
  Here are some lemmas that may be handy in calculation.
-------------------------------------------------------------------------%

let CUP_CLa_CUP_less = prove_thm
  (`CUP_CLa_CUP_less`,
   "!r X (x:*). 
        CUP_CLa r /\ ~(X={}) ==> (IS_UB r X x = r (CUP r X) x)",
    REWRITE_TAC [CUP_CLa] THEN REPEAT STRIP_TAC 
    THEN RES_TAC THEN IMP_RES_TAC CUP_less 
    THEN ASM_REWRITE_TAC[]) ;;
      
let CUP_PLa_CUP_less = prove_thm
  (`CUP_PLa_CUP_less`,
   "!r X (x:*). 
        CUP_PLa r ==> (IS_UB r X x = r (CUP r X) x)",
    REWRITE_TAC [CUP_PLa] THEN REPEAT STRIP_TAC 
    THEN XRULE_ASSUM_TAC (SPEC "X:(*)set") THEN UNDISCH_ALL_TAC
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CUP_less 
    THEN ASM_REWRITE_TAC[]) ;;

let CUP_CLa_CUP_greater = prove_thm
  (`CUP_CLa_CUP_greater`,
   "!r X (x:*). CUP_CLa r /\ x IN X ==> r x (CUP r X)",
    REWRITE_TAC [CUP_CLa] THEN REPEAT STRIP_TAC 
    THEN ASM_CASES_TAC "(X:(*)set)={}"
    THENL
    [  UNDISCH_TAC "(x:*) IN X"
       THEN ASM_REWRITE_TAC [NOT_IN_EMPTY] ;
       RES_TAC THEN IMP_RES_TAC CUP_greater ]) ;;

let CUP_PLa_CUP_greater = prove_thm
  (`CUP_PLa_CUP_greater`,
   "!r X (x:*). CUP_PLa r /\ x IN X ==> r x (CUP r X)",
    REPEAT STRIP_TAC 
    THEN IMP_RES_TAC CUP_PLa_IMP_CUP_CLa
    THEN IMP_RES_TAC CUP_CLa_CUP_greater) ;;

let CAP_CLa_CAP_greater = prove_thm
  (`CAP_CLa_CAP_greater`,
   "!r X (x:*). 
        CAP_CLa r /\ ~(X={}) ==> (IS_LB r X x = r x (CAP r X))",
    REWRITE_TAC [CAP_CLa] THEN REPEAT STRIP_TAC 
    THEN RES_TAC THEN IMP_RES_TAC CAP_greater
    THEN ASM_REWRITE_TAC[]) ;;

let CAP_PLa_CAP_greater = prove_thm
  (`CAP_PLa_CAP_greater`,
   "!r X (x:*). 
        CAP_PLa r ==> (IS_LB r X x = r x (CAP r X))",
    REWRITE_TAC [CAP_PLa] THEN REPEAT STRIP_TAC 
    THEN XRULE_ASSUM_TAC (SPEC "X:(*)set") THEN UNDISCH_ALL_TAC
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CAP_greater
    THEN ASM_REWRITE_TAC[]) ;;
      
let CAP_CLa_CAP_less = prove_thm
  (`CAP_CLa_CAP_less`,
   "!r X (x:*). CAP_CLa r /\ x IN X ==> r (CAP r X) x",
    REWRITE_TAC [CAP_CLa] THEN REPEAT STRIP_TAC 
    THEN ASM_CASES_TAC "(X:(*)set)={}"
    THENL
    [  UNDISCH_TAC "(x:*) IN X"
       THEN ASM_REWRITE_TAC [NOT_IN_EMPTY] ;
       RES_TAC THEN IMP_RES_TAC CAP_less ]) ;;

let CAP_PLa_CAP_less = prove_thm
  (`CAP_PLa_CAP_less`,
   "!r X (x:*). CAP_PLa r /\ x IN X ==> r (CAP r X) x",
    REPEAT STRIP_TAC 
    THEN IMP_RES_TAC CAP_PLa_IMP_CAP_CLa
    THEN IMP_RES_TAC CAP_CLa_CAP_less) ;;

let CLa_CUP_less = prove_thm
  (`CLa_CUP_less`,
   "!r X (x:*). 
        CLa r  ==> (IS_UB r X x = r (CUP r X) x)",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC 
    THEN IMP_RES_TAC CUP_PLa_CUP_less
    THEN ASM_REWRITE_TAC[]) ;;

let CLa_CUP_greater = prove_thm
  (`CLa_CUP_greater`,
   "!r X (x:*). CLa r /\ x IN X ==> r x (CUP r X)",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC 
    THEN IMP_RES_TAC CUP_PLa_CUP_greater) ;;

let CLa_CAP_greater = prove_thm
  (`CLa_CAP_greater`,
   "!r X (x:*). CLa r  ==> (IS_LB r X x = r x (CAP r X))",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC 
    THEN IMP_RES_TAC CAP_PLa_CAP_greater
    THEN ASM_REWRITE_TAC[]) ;;

let CLa_CAP_less = prove_thm
  (`CLa_CAP_less`,
   "!r X (x:*). CLa r /\ x IN X ==> r (CAP r X) x",
    REWRITE_TAC [CLa] THEN REPEAT STRIP_TAC 
    THEN IMP_RES_TAC CAP_PLa_CAP_less) ;;

%-------------------------------------------------------------------------
       A pointed CAP lattice defines a CUP lattice, and vice versa
-------------------------------------------------------------------------%

let CAP_PLa_EQU_CUP_PLa = prove_thm
  (`CAP_PLa_EQU_CUP_PLa`,
   "CAP_PLa (r:*->*->bool) = CUP_PLa r",
    EQ_TAC
    THENL
    [ %-- CAP implies CUP --%
      REWRITE_TAC [CUP_PLa]
      THEN REPEAT STRIP_TAC  
      THENL
      [ RULE_ASSUM_TAC (REWRITE_RULE [CAP_PLa]) THEN ASM_REWRITE_TAC[] ;
        EXISTS_TAC "CAP r {a:* | (!x:*. x IN X ==> r x a)}"
        THEN REWRITE_TAC [isCUP; IS_UB]
        THEN REPEAT STRIP_TAC
        THENL
        [ IMP_RES_TAC CAP_PLa_CAP_greater
          THEN FIRST_ASSUM MATCH_MP_TAC
          THEN REWRITE_TAC [IS_LB] 
          THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
          THEN REPEAT STRIP_TAC THEN RES_TAC ;
          MATCH_MP_TAC CAP_PLa_CAP_less
          THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
          THEN ASM_REWRITE_TAC[] 
          ] ] ;
      %-- CUP implies CAP --%
      REWRITE_TAC [CAP_PLa]
      THEN REPEAT STRIP_TAC  
      THENL
      [ RULE_ASSUM_TAC (REWRITE_RULE [CUP_PLa]) THEN ASM_REWRITE_TAC[] ;
        EXISTS_TAC "CUP r {a:* | (!x:*. x IN X ==> r a x)}"
        THEN REWRITE_TAC [isCAP; IS_LB]
        THEN REPEAT STRIP_TAC
        THENL
        [ IMP_RES_TAC CUP_PLa_CUP_less
          THEN FIRST_ASSUM MATCH_MP_TAC
          THEN REWRITE_TAC [IS_UB] 
          THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
          THEN REPEAT STRIP_TAC THEN RES_TAC ;
          MATCH_MP_TAC CUP_PLa_CUP_greater
          THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
          THEN ASM_REWRITE_TAC[] 
          ] ] 
     ]) ;;

%-------------------------------------------------------------------------
    Consequently, a complete lattice is simply a CAP (or CUP) lattice
-------------------------------------------------------------------------%

let CLa_EQU_CAP_PLa = prove_thm
  (`CLa_EQU_CAP_PLa`,
   "CLa (r:*->*->bool) = CAP_PLa r",
    REWRITE_TAC [CLa; CAP_PLa_EQU_CUP_PLa]) ;;

%-------------------------------------------------------------------------
                A pointed CUP lattice is a CPO.
-------------------------------------------------------------------------%

let CUP_PLa_IMP_CPO = prove_thm
  (`CUP_PLa_IMP_CPO`,
   "!r:*->*->bool. CUP_PLa r  ==> CPO r",
    REWRITE_TAC [CUP_PLa; CPO]
    THEN REPEAT STRIP_TAC 
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
  Here is the definition of CAP and CUP distributive/junctive functions.
-------------------------------------------------------------------------%

let CUP_Distr = new_definition
  (`CUP_Distr`,
   "CUP_Distr r s (f:*->**) = (!X. f (CUP r X) = CUP s (IMAGE f X))") ;;

let CAP_Distr = new_definition
  (`CAP_Distr`,
   "CAP_Distr r s (f:*->**) = (!X. f (CAP r X) = CAP s (IMAGE f X))") ;;

let CUP_Junct = new_definition
  (`CUP_Junct`,
   "CUP_Junct r s (f:*->**) = 
        (!X. ~(X={}) ==> (f (CUP r X) = CUP s (IMAGE f X)))") ;;

let CAP_Junct = new_definition
  (`CAP_Junct`,
   "CAP_Junct r s (f:*->**) = 
        (!X. ~(X={}) ==> (f (CAP r X) = CAP s (IMAGE f X)))") ;;

%-------------------------------------------------------------------------
  Clearly a distributive function is also junctive.
-------------------------------------------------------------------------%

let CUP_Distr_IMP_Junct = prove_thm
  (`CUP_Distr_IMP_Junct`,
   "!r s (f:*->**). CUP_Distr r s f ==> CUP_Junct r s f",
    REWRITE_TAC [CUP_Distr; CUP_Junct] 
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]) ;;

let CAP_Distr_IMP_Junct = prove_thm
  (`CAP_Distr_IMP_Junct`,
   "!r s (f:*->**). CAP_Distr r s f ==> CAP_Junct r s f",
    REWRITE_TAC [CAP_Distr; CAP_Junct] 
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
  For a CUP-distributive function f we have f Bot = Bot. (Consequently Bot 
  is the least fix point of f).
  Analog for CAP.
-------------------------------------------------------------------------%

let CUP_Distr_Bot = prove_thm
  (`CUP_Distr_Bot`,
   "!r s (f:*->**). 
       CLa r /\ CLa s /\ CUP_Distr r s f ==> (f (Bot r) = (Bot s))",
    REWRITE_TAC [CUP_Distr; CLa; CAP_PLa; PO]
    THEN REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN "Bot (r:*->*->bool) = CUP r {}" SUBST1_TAC
    THENL 
    [ MATCH_MP_TAC Bot_EQU_CUP_EMPTY THEN ASM_REWRITE_TAC[] ;
      ALL_TAC ]
    THEN SUBGOAL_THEN "Bot (s:**->**->bool) = CUP s {}" SUBST1_TAC
    THENL 
    [ MATCH_MP_TAC Bot_EQU_CUP_EMPTY THEN ASM_REWRITE_TAC[] ;
      ALL_TAC ]
    THEN ASM_REWRITE_TAC[IMAGE_EMPTY]) ;;

let CAP_Distr_Top = prove_thm
  (`CAP_Distr_Top`,
   "!r s (f:*->**). 
       CLa r /\ CLa s /\ CAP_Distr r s f ==> (f (Top r) = (Top s))",
    REWRITE_TAC [CAP_Distr; CLa; CUP_PLa; PO]
    THEN REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN "Top (r:*->*->bool) = CAP r {}" SUBST1_TAC
    THENL 
    [ MATCH_MP_TAC Top_EQU_CAP_EMPTY THEN ASM_REWRITE_TAC[] ;
      ALL_TAC ]
    THEN SUBGOAL_THEN "Top (s:**->**->bool) = CAP s {}" SUBST1_TAC
    THENL 
    [ MATCH_MP_TAC Top_EQU_CAP_EMPTY THEN ASM_REWRITE_TAC[] ;
      ALL_TAC ]
    THEN ASM_REWRITE_TAC[IMAGE_EMPTY]) ;;

%-------------------------------------------------------------------------
  A CUP-junctive function that maps to a complete lattice is continous.
-------------------------------------------------------------------------%

let CUP_Junct_IMP_Cont = prove_thm
  (`CUP_Junct_IMP_Cont`,
   "!r s (f:*->**). CLa s /\ CUP_Junct r s f ==> Cont r s f",
    REWRITE_TAC [CUP_Junct; CLa; CUP_PLa; Cont] 
    THEN REPEAT STRIP_TAC
    THENL
    [ REWRITE_TAC [Directed] THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[] ;
      RES_TAC ]) ;;

%-------------------------------------------------------------------------
              A CAP-junctive function is monotonic
-------------------------------------------------------------------------%

let CAP_Junct_IMP_Mono = prove_thm
  (`CAP_Junct_IMP_Mono`,
   "CAP_PLa r /\ CAP_PLa s /\ CAP_Junct r s (f:*->**) ==> Mono r s f",
    REWRITE_TAC [CAP_Junct; Mono] THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN "TRANS (s:**->**->bool)" 
         (\thm. MATCH_MP_TAC (REWRITE_RULE [TRANS_DEF ] thm))
    THENL [RULE_ASSUM_TAC (REWRITE_RULE [CAP_PLa; PO])
           THEN ASM_REWRITE_TAC []  ; ALL_TAC ]
    THEN EXISTS_TAC "CAP s (IMAGE (f:*->**) {x,y})"
    THEN CONJ_TAC THENL
    [ %-- 1 --%
      SUBGOAL_THEN "~({x:*,y} = {})" ASSUME_TAC
      THENL [ REWRITE_TAC [SYM (SPEC_ALL MEMBER_NOT_EMPTY)]
              THEN EXISTS_TAC "x:*" THEN REWRITE_TAC [IN_INSERT] ; ALL_TAC ]
      THEN RES_TAC
      THEN FIRST_ASSUM (SUBST1_TAC o SYM)
      THEN SUBGOAL_THEN "CAP r {x:*,y} = x" SUBST1_TAC
      THENL [ MATCH_MP_TAC isCAP_CAP 
              THEN REWRITE_TAC [isCAP; IS_LB; IN_INSERT; NOT_IN_EMPTY]
              THEN RULE_ASSUM_TAC (REWRITE_RULE [CAP_PLa; PO; REFL_DEF])
              THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
              THEN FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[] ;
              ALL_TAC ]
      THEN RULE_ASSUM_TAC (REWRITE_RULE [CAP_PLa; PO; REFL_DEF])
      THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      MATCH_MP_TAC CAP_PLa_CAP_less
      THEN ASM_REWRITE_TAC [IN_IMAGE]
      THEN EXISTS_TAC "y:*" THEN REWRITE_TAC [IN_INSERT]
     ]) ;;

%-------------------------------------------------------------------------
  Here begins the proof of Knaster-Tarski finx point theorem for 
  monotonic functions.
-------------------------------------------------------------------------%

let a = "CAP r {x:* | r (f x) x}" ;;
let A = "{x | r ((f:*->*) x) x}" ;;

%-------------------------------------------------------------------------
  We are going to show that a is a fix point. First, show that 
  f a is less than a.
-------------------------------------------------------------------------%

let lemma1 = prove
  ("!r (f:*->*). CAP_PLa r /\ Mono r r f ==> r (f ^a) ^a",
    REWRITE_TAC [Mono] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC CAP_PLa_CAP_greater
    THEN EVERY_ASSUM 
         (\thm. REWRITE_TAC [SYM (SPEC_ALL thm)] ? ALL_TAC)
    THEN REWRITE_TAC[IS_LB]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV   
    THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN "TRANS (r:*->*->bool)" ASSUME_TAC
    THENL [ RULE_ASSUM_TAC (REWRITE_RULE [CAP_PLa; PO])
            THEN ASM_REWRITE_TAC[] ; ALL_TAC ]
    THEN FIRST_ASSUM (\thm. MATCH_MP_TAC (REWRITE_RULE [TRANS_DEF] thm))
    THEN EXISTS_TAC "(f:*->*) b" THEN ASM_REWRITE_TAC[]
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN MATCH_MP_TAC CAP_PLa_CAP_less
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV 
    THEN ASM_REWRITE_TAC[]) ;;
    
%-------------------------------------------------------------------------
   Consequently f a is an element of A.
-------------------------------------------------------------------------%

let lemma2 = prove
  ("!r (f:*->*). CAP_PLa r /\ Mono r r f ==> (f ^a) IN ^A",
    (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC lemma1
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [Mono])
    THEN RES_TAC ) ;;

%-------------------------------------------------------------------------
   Since a is the CAP of A and f a IN A, it follows that a is less than
   f a.
-------------------------------------------------------------------------%

let lemma3 = prove
  ("!r (f:*->*). CAP_PLa r /\ Mono r r f ==> r ^a (f ^a)",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC lemma2
    THEN MATCH_MP_TAC CAP_PLa_CAP_less
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
   It follows from the anti-symmetry of r that f a = a.
-------------------------------------------------------------------------%

let lemma4 = prove
  ("!r (f:*->*). CAP_PLa r /\ Mono r r f ==> ((f ^a) = ^a)",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC lemma1
    THEN IMP_RES_TAC lemma3
    THEN UNDISCH_ALL_TAC THEN REWRITE_TAC [CAP_PLa; PO; ANTI_SYM]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;;

%-------------------------------------------------------------------------
   From the reflexivity of r it follows that a is also less than any
   fix point.
-------------------------------------------------------------------------%

let lemma5 = prove
  ("!r (f:*->*) x. CAP_PLa r /\ Mono r r f /\ (f x = x) 
                  ==> r ^a x",
    REWRITE_TAC [CAP_PLa; PO; REFL_DEF]
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC A)
    THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CAP_less
    THEN EXISTS_TAC "b:*" 
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
   Here are the Knaster-Tarski theorem of fix point of a monotonic 
   function.
-------------------------------------------------------------------------%

let TARSKI_MONO_isLFP = prove_thm
  (`TARSKI_MONO_isLFP`,
   "!r (f:*->*). 
        CAP_PLa r /\ Mono r r f  ==> isLFP r f (CAP r {x | r (f x) x})",
   REWRITE_TAC [isLFP]
   THEN REPEAT STRIP_TAC
   THENL [ IMP_RES_TAC lemma4 ; IMP_RES_TAC lemma5 ]) ;;

let TARSKI_MONO_LFP = prove_thm
  (`TARSKI_MONO_LFP`,
   "!r (f:*->*). 
        CAP_PLa r /\ Mono r r f 
        ==>
        (LFP r f = CAP r {x | r (f x) x})",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC UNIQUE_LFP
    THEN EXISTS_TAC "r:*->*->bool"
    THEN EXISTS_TAC "f:*->*"
    THEN IMP_RES_TAC TARSKI_MONO_isLFP
    THEN UNDISCH_ALL_TAC 
    THEN REWRITE_TAC [CAP_PLa; PO] THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]
    THEN IMP_RES_TAC LFP) ;;

let TARSKI_MONO_equ = prove_thm
  (`TARSKI_MONO_equ`,
   "!r (f:*->*). 
        CAP_PLa r /\ Mono r r f  ==>  (f (LFP r f) = LFP r f)",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC TARSKI_MONO_LFP
    THEN IMP_RES_TAC (REWRITE_RULE [isLFP] TARSKI_MONO_isLFP)
    THEN ASM_REWRITE_TAC[]) ;;

let TARSKI_MONO_less = prove_thm
  (`TARSKI_MONO_less`,
   "!r (f:*->*) x. 
        CAP_PLa r /\ Mono r r f /\ (r (f x) x)
        ==>
        r (LFP r f) x",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC TARSKI_MONO_LFP
    THEN ASM_REWRITE_TAC[]
    THEN MATCH_MP_TAC CAP_less
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [CAP_PLa])
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------
   Here is the Rolling Rule.
-------------------------------------------------------------------------%

let lemma1 = prove
  ("!r s (f:*->**) (g:**->*). 
        CAP_PLa r /\ Mono r s f /\ 
        CAP_PLa s /\ Mono s r g
        ==>
        s (f (LFP r (g o f))) (LFP s (f o g))",
    REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN "Mono s s ((f:*->**) o (g:**->*))" ASSUME_TAC
    THENL [ IMP_RES_TAC Mono_o ; ALL_TAC ]
    THEN IMP_RES_TAC TARSKI_MONO_equ
    THEN EVERY_ASSUM (\thm. SUBST1_TAC (SYM thm) ? ALL_TAC)
    THEN REWRITE_TAC [o_THM]
    THEN MATCH_MP_TAC (REWRITE_RULE [Mono](ASSUME "Mono r s (f:*->**)"))
    THEN MATCH_MP_TAC TARSKI_MONO_less
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [o_THM])
    THEN ASM_REWRITE_TAC [o_THM]
    THEN CONJ_TAC 
    THENL [ IMP_RES_TAC Mono_o ; ALL_TAC ]
    THEN UNDISCH_ALL_TAC
    THEN REWRITE_TAC [ CAP_PLa; PO; REFL_DEF] THEN STRIP_TAC
    THEN ASM_REWRITE_TAC[]) ;;

let lemma2 = prove
  ("!r s (f:*->**) (g:**->*). 
        CAP_PLa r /\ Mono r s f /\ 
        CAP_PLa s /\ Mono s r g
        ==>
        s (LFP s (f o g)) (f (LFP r (g o f))) ",
    REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN "Mono r r ((g:**->*) o (f:*->**))" ASSUME_TAC
    THENL [ IMP_RES_TAC Mono_o ; ALL_TAC ]
    THEN IMP_RES_TAC TARSKI_MONO_equ
    THEN SUBGOAL_THEN "Mono s s ((f:*->**) o (g:**->*))" ASSUME_TAC
    THENL [ IMP_RES_TAC Mono_o ; ALL_TAC ]
    THEN MATCH_MP_TAC TARSKI_MONO_less 
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [o_THM])
    THEN ASM_REWRITE_TAC [o_THM]
    THEN UNDISCH_ALL_TAC
    THEN REWRITE_TAC [CAP_PLa; PO; REFL_DEF] THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]) ;;

let ROLING_THM = prove_thm
  (`ROLING_THM`,
   "!r s (f:*->**) (g:**->*). 
        CAP_PLa r /\ Mono r s f /\ 
        CAP_PLa s /\ Mono s r g
        ==>
        (f (LFP r (g o f)) = LFP s (f o g))",
   REPEAT STRIP_TAC 
   THEN SUBGOAL_THEN "ANTI_SYM (s:**->**->bool)" 
        (\thm. MATCH_MP_TAC (REWRITE_RULE [ANTI_SYM ]thm))
   THENL [ UNDISCH_ALL_TAC 
           THEN REWRITE_TAC [CAP_PLa; PO] THEN TAUT_TAC ;
           ALL_TAC ]
   THEN CONJ_TAC
   THENL [ MATCH_MP_TAC lemma1 THEN ASM_REWRITE_TAC[] ;
           MATCH_MP_TAC lemma2 THEN ASM_REWRITE_TAC[]] ) ;;

%-------------------------------------------------------------------------
   Here is Plotkin's lemma of Fix Point Fusion.
-------------------------------------------------------------------------%

let FP_lower_FUSION = prove_thm
  (`FP_lower_FUSION`,
   "!r s (f:*->**) (g:*->*) (h:**->**). 
        CAP_PLa r /\ Mono r r g /\ 
        CAP_PLa s /\ Mono s s h /\
        (!x. s ((h o f) x) ((f o g) x))
        ==>
        s (LFP s h) (f (LFP r g))",
   REWRITE_TAC [o_THM] THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC TARSKI_MONO_less
   THEN IMP_RES_TAC TARSKI_MONO_equ
   THEN ASM_REWRITE_TAC[]
   THEN UNDISCH_ALL_TAC 
   THEN REWRITE_TAC [CAP_PLa; PO; REFL_DEF; TRANS_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "(f:*->**) (g (LFP r g))"
   THEN ASM_REWRITE_TAC[]) ;;

close_theory() ;;

